# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various loss functions for policy gradients.
"""

from typing import Optional

import torch
import torch.nn.functional as F

from texar.torch.losses.losses_utils import mask_and_reduce
from texar.torch.utils.shapes import get_rank

__all__ = [
    "pg_loss_with_logits",
    "pg_loss_with_log_probs",
]


def pg_loss_with_logits(actions: torch.Tensor,
                        logits: torch.Tensor,
                        advantages: torch.Tensor,
                        rank: Optional[int] = None,
                        batched: bool = False,
                        sequence_length: Optional[torch.LongTensor] = None,
                        average_across_batch: bool = True,
                        average_across_timesteps: bool = False,
                        average_across_remaining: bool = False,
                        sum_over_batch: bool = False,
                        sum_over_timesteps: bool = True,
                        sum_over_remaining: bool = True,
                        time_major: bool = False) -> torch.Tensor:
    r"""Policy gradient loss with logits. Used for discrete actions.

    `pg_loss = reduce( advantages * -log_prob( actions )  )`,
    where `advantages` and `actions` do not back-propagate gradients.

    All arguments except :attr:`logits` and :attr:`actions` are the same with
    :func:`pg_loss_with_log_probs`.

    Args:
        actions: Tensor of shape
            `[(batch_size,) max_time, d_3, ..., d_rank]` and of dtype
            `int32` or `int64`.
            The rank of the Tensor is specified with :attr:`rank`.

            The batch dimension exists only if :attr:`batched` is `True`.

            The batch and time dimensions
            are exchanged, i.e., `[max_time, batch_size, ...]` if
            :attr:`time_major` is `True`.
        logits: Unscaled log probabilities of shape
            `[(batch_size,) max_time, d_3, ..., d_{rank+1}]`
            and dtype `float32` or `float64`.
            The batch and time dimensions are exchanged if `time_major`
            is `True`.
        advantages: Tensor of shape
            `[(batch_size,) max_time, d_3, ..., d_rank]` and
            dtype `float32` or `float64`.
            The batch and time dimensions are exchanged if `time_major`
            is `True`.
        rank (int, optional): The rank of :attr:`actions`.
            If `None` (default), rank is automatically inferred from
            `actions` or `advantages`. If the inference fails,
            `rank` is set to 1 if :attr:`batched` is `False`,
            and set to 2 if :attr:`batched` is `True`.
        batched (bool): `True` if the inputs are batched.
        sequence_length (optional): A Tensor of shape `[batch_size]`.
            Time steps beyond the respective sequence lengths will have zero
            losses. Used if :attr:`batched` is `True`.
        average_across_timesteps (bool): If set, average the loss across
            the time dimension. Must not set `average_across_timesteps`
            and `sum_over_timesteps` at the same time.
        average_across_batch (bool): If set, average the loss across the
            batch dimension. Must not set `average_across_batch`'
            and `sum_over_batch` at the same time.
            Ignored if `batched` is `False`.
        average_across_remaining (bool): If set, average the sequence across the
            remaining dimensions. Must not set `average_across_remaining`'
            and `sum_over_remaining` at the same time. Ignored if
            no more dimensions other than the batch and time dimensions.
        sum_over_timesteps (bool): If set, sum the loss across the
            time dimension. Must not set `average_across_timesteps`
            and `sum_over_timesteps` at the same time.
        sum_over_batch (bool): If set, sum the loss across the
            batch dimension. Must not set `average_across_batch`
            and `sum_over_batch` at the same time.
            Ignored if `batched` is `False`.
        sum_over_remaining (bool): If set, sum the loss across the
            remaining dimension. Must not set `average_across_remaining`
            and `sum_over_remaining` at the same time. Ignored if
            no more dimensions other than the batch and time dimensions.
        time_major (bool): The shape format of the inputs. If `True`,
            :attr:`logits`, :attr:`actions` and :attr:`advantages` must
            have shape `[max_time, batch_size, ...]`. If `False` (default),
            they must have shape `[batch_size, max_time, ...]`.
            Ignored if `batched` is `False`.

    Returns:
        A Tensor containing the loss to minimize, whose rank depends on the
        reduce arguments. For example, the batch dimension is reduced if
        either :attr:`average_across_batch` or :attr:`sum_over_batch` is
        `True`, which decreases the rank of output tensor by 1.
    """
    actions = actions.detach()
    logits = F.log_softmax(logits, dim=-1)
    logits = logits.permute((0, -1) + tuple(range(1, logits.dim() - 1)))
    neg_log_probs = F.nll_loss(logits, actions, reduction='none')

    return pg_loss_with_log_probs(
        log_probs=-neg_log_probs,
        advantages=advantages,
        rank=rank,
        batched=batched,
        sequence_length=sequence_length,
        average_across_batch=average_across_batch,
        average_across_timesteps=average_across_timesteps,
        average_across_remaining=average_across_remaining,
        sum_over_batch=sum_over_batch,
        sum_over_timesteps=sum_over_timesteps,
        sum_over_remaining=sum_over_remaining,
        time_major=time_major)


def pg_loss_with_log_probs(log_probs: torch.Tensor,
                           advantages: torch.Tensor,
                           rank: Optional[int] = None,
                           batched: bool = False,
                           sequence_length: Optional[torch.LongTensor] = None,
                           average_across_batch: bool = True,
                           average_across_timesteps: bool = False,
                           average_across_remaining: bool = False,
                           sum_over_batch: bool = False,
                           sum_over_timesteps: bool = True,
                           sum_over_remaining: bool = True,
                           time_major: bool = False) -> torch.Tensor:
    r"""Policy gradient loss with log probabilities of actions.

    `pg_loss = reduce(advantages * -log_probs)`,
    where `advantages` does not back-propagate gradients.

    All arguments except :attr:`log_probs` are the same as
    :func:`pg_loss_with_logits`.

    Args:
        log_probs: Log probabilities of shape
            `[(batch_size,) max_time, ..., d_rank]` and dtype `float32`
            or `float64`. The rank of the Tensor is specified
            with :attr:`rank`.

            The batch dimension exists only if :attr:`batched` is `True`.

            The batch and time dimensions are exchanged, i.e.,
            `[max_time, batch_size, ...]` if :attr:`time_major` is `True`.
        advantages: Tensor of shape
            `[(batch_size,) max_time, d_3, ..., d_rank]` and
            dtype `float32` or `float64`.
            The batch dimension exists only if `batched` is `True`.
            The batch and time dimensions
            are exchanged if `time_major` is `True`.
        rank (int, optional): The rank of :attr:`log_probs`.
            If `None` (default), rank is automatically inferred from
            `log_probs` or `advantages`. If the inference fails,
            `rank` is set to 1 if `batched``==False`,
            and set to 2 if `batched``==True`.
        batched (bool): `True` if the inputs are batched.
        sequence_length (optional): A Tensor of shape `[batch_size]`.
            Time steps beyond the respective sequence lengths will have zero
            losses. Used if :attr:`batched` is `True`.
        average_across_timesteps (bool): If set, average the loss across
            the time dimension. Must not set `average_across_timesteps`
            and `sum_over_timesteps` at the same time.
        average_across_batch (bool): If set, average the loss across the
            batch dimension. Must not set `average_across_batch`'
            and `sum_over_batch` at the same time.
            Ignored if `batched` is `False`.
        average_across_remaining (bool): If set, average the sequence across the
            remaining dimensions. Must not set `average_across_remaining`'
            and `sum_over_remaining` at the same time. Ignored if
            no more dimensions other than the batch and time dimensions.
        sum_over_timesteps (bool): If set, sum the loss across the
            time dimension. Must not set `average_across_timesteps`
            and `sum_over_timesteps` at the same time.
        sum_over_batch (bool): If set, sum the loss across the
            batch dimension. Must not set `average_across_batch`
            and `sum_over_batch` at the same time.
            Ignored if `batched` is `False`.
        sum_over_remaining (bool): If set, sum the loss across the
            remaining dimension. Must not set `average_across_remaining`
            and `sum_over_remaining` at the same time. Ignored if
            no more dimensions other than the batch and time dimensions.
        time_major (bool): The shape format of the inputs. If `True`,
            :attr:`log_probs` and :attr:`advantages` must have shape
            `[max_time, batch_size, ...]`. If `False` (default),
            they must have shape `[batch_size, max_time, ...]`.
            Ignored if :attr:`batched` is `False`.

    Returns:
        A Tensor containing the loss to minimize, whose rank depends on the
        reduce arguments. For example, the batch dimension is reduced if
        either :attr:`average_across_batch` or :attr:`sum_over_batch` is
        `True`, which decreases the rank of output tensor by 1.
    """
    advantages = advantages.detach()

    losses = -log_probs * advantages

    if rank is None:
        rank = get_rank(log_probs) or get_rank(advantages)
    if rank is None:
        rank = 2 if batched else 1

    if batched:
        losses = mask_and_reduce(
            losses,
            sequence_length,
            rank=rank,
            average_across_batch=average_across_batch,
            average_across_timesteps=average_across_timesteps,
            average_across_remaining=average_across_remaining,
            sum_over_batch=sum_over_batch,
            sum_over_timesteps=sum_over_timesteps,
            sum_over_remaining=sum_over_remaining,
            time_major=time_major)
    elif rank > 1:
        if average_across_remaining and sum_over_remaining:
            raise ValueError("Only one of `average_across_remaining` and "
                             "`sum_over_remaining` can be set.")
        if average_across_remaining:
            for average_axis in sorted(list(range(1, rank)), reverse=True):
                losses = torch.mean(losses, dim=average_axis)
        elif sum_over_remaining:
            for sum_axis in sorted(list(range(1, rank)), reverse=True):
                losses = torch.sum(losses, dim=sum_axis)

    if not batched:
        if average_across_timesteps and sum_over_timesteps:
            raise ValueError("Only one of `average_across_timesteps` and "
                             "`sum_over_timesteps` can be set.")
        if average_across_timesteps:
            losses = torch.mean(losses, dim=0)
        elif sum_over_timesteps:
            losses = torch.sum(losses, dim=0)

    return losses



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for various optimization related utilities.
"""

import unittest

import torch

from texar.torch.core.optimization import *


class OptimizationTest(unittest.TestCase):
    r"""Test optimization.
    """

    def setUp(self):
        N, D_in, H, D_out = 64, 100, 10, 1

        self.x = torch.randn(N, D_in)
        self.y = torch.randn(N, D_out)

        self.model = torch.nn.Sequential(
            torch.nn.Linear(D_in, H),
            torch.nn.ReLU(),
            torch.nn.Linear(H, D_out),)

        self.loss_fn = torch.nn.MSELoss(reduction='sum')

    def test_get_optimizer(self):
        r"""Tests get_optimizer.
        """
        default_optimizer = get_optimizer(params=[torch.tensor(1)],
                                          hparams=None)
        self.assertIsInstance(default_optimizer, torch.optim.Adam)

        hparams = {
            "optimizer": {
                "type": "RMSprop",
                "kwargs": {
                    "lr": 0.001,
                    "alpha": 0.99,
                    "eps": 1e-8,
                    "weight_decay": 0,
                    "momentum": 0,
                    "centered": False
                }
            },
            "learning_rate_decay": {
                "type": "",
                "kwargs": {}
            },
            "gradient_clip": {
                "type": "",
                "kwargs": {}
            },
            "gradient_noise_scale": None,
            "name": None
        }

        rmsprop_optimizer = get_optimizer(params=[torch.tensor(1)],
                                          hparams=hparams)
        self.assertIsInstance(rmsprop_optimizer, torch.optim.RMSprop)

        hparams = {
            "optimizer": {
                "type": torch.optim.SGD,
                "kwargs": {
                    "lr": 0.001,
                    "weight_decay": 0,
                    "momentum": 0
                }
            },
            "learning_rate_decay": {
                "type": "",
                "kwargs": {}
            },
            "gradient_clip": {
                "type": "",
                "kwargs": {}
            },
            "gradient_noise_scale": None,
            "name": None
        }

        sgd_optimizer = get_optimizer(params=[torch.tensor(1)],
                                      hparams=hparams)
        self.assertIsInstance(sgd_optimizer, torch.optim.SGD)

    def test_get_scheduler(self):
        r"""Tests get_scheduler.
        """
        optimizer = get_optimizer(params=[torch.tensor(1)], hparams=None)

        default_scheduler = get_scheduler(optimizer=optimizer,
                                          hparams=None)
        self.assertEqual(default_scheduler, None)

        hparams = {
            "optimizer": {
                "type": "",
                "kwargs": {}
            },
            "learning_rate_decay": {
                "type": "ExponentialLR",
                "kwargs": {
                    "gamma": 0.99
                }
            },
            "gradient_clip": {
                "type": "",
                "kwargs": {}
            },
            "gradient_noise_scale": None,
            "name": None
        }

        scheduler = get_scheduler(optimizer=optimizer,
                                  hparams=hparams)
        self.assertIsInstance(scheduler, torch.optim.lr_scheduler.ExponentialLR)

        hparams = {
            "optimizer": {
                "type": "",
                "kwargs": {}
            },
            "learning_rate_decay": {
                "type": torch.optim.lr_scheduler.ExponentialLR,
                "kwargs": {
                    "gamma": 0.99
                }
            },
            "gradient_clip": {
                "type": "",
                "kwargs": {}
            },
            "gradient_noise_scale": None,
            "name": None
        }

        scheduler = get_scheduler(optimizer=optimizer,
                                  hparams=hparams)
        self.assertIsInstance(scheduler, torch.optim.lr_scheduler.ExponentialLR)

    def test_get_grad_clip_fn(self):
        r"""Tests get_grad_clip_fn.
        """
        default_grad_clip_fn = get_grad_clip_fn(hparams=None)
        self.assertEqual(default_grad_clip_fn, None)

        hparams = {
            "optimizer": {
                "type": "",
                "kwargs": {}
            },
            "learning_rate_decay": {
                "type": "",
                "kwargs": {}
            },
            "gradient_clip": {
                "type": "clip_grad_norm_",
                "kwargs": {
                    "max_norm": 10,
                    "norm_type": 2
                }
            },
            "gradient_noise_scale": None,
            "name": None
        }

        grad_clip_fn = get_grad_clip_fn(hparams=hparams)
        if not callable(grad_clip_fn):
            raise ValueError("grad_clip_fn is not callable")

        hparams = {
            "optimizer": {
                "type": "",
                "kwargs": {}
            },
            "learning_rate_decay": {
                "type": "",
                "kwargs": {}
            },
            "gradient_clip": {
                "type": torch.nn.utils.clip_grad_norm_,
                "kwargs": {
                    "max_norm": 10,
                    "norm_type": 2
                }
            },
            "gradient_noise_scale": None,
            "name": None
        }

        grad_clip_fn = get_grad_clip_fn(hparams=hparams)
        if not callable(grad_clip_fn):
            raise ValueError("grad_clip_fn is not callable")

    def test_get_train_op(self):
        r"""Tests get_train_op.
        """
        hparams = {
            "optimizer": {
                "type": torch.optim.SGD,
                "kwargs": {
                    "lr": 0.001
                }
            },
            "learning_rate_decay": {
                "type": torch.optim.lr_scheduler.ExponentialLR,
                "kwargs": {
                    "gamma": 0.99
                }
            },
            "gradient_clip": {
                "type": torch.nn.utils.clip_grad_norm_,
                "kwargs": {
                    "max_norm": 10,
                    "norm_type": 2
                }
            },
            "gradient_noise_scale": None,
            "name": None
        }

        # Case 1
        optimizer = get_optimizer(self.model.parameters(), hparams)
        train_op = get_train_op(optimizer=optimizer, hparams=hparams)

        for t in range(50):
            y_pred = self.model(self.x)
            loss = self.loss_fn(y_pred, self.y)
            loss.backward()
            train_op()

        # Case 2
        train_op = get_train_op(params=self.model.parameters(), hparams=hparams)

        for t in range(50):
            y_pred = self.model(self.x)
            loss = self.loss_fn(y_pred, self.y)
            loss.backward()
            train_op()

        # Case 3
        optimizer = get_optimizer(self.model.parameters(), hparams)
        scheduler = get_scheduler(optimizer=optimizer,
                                  hparams=hparams)
        train_op = get_train_op(scheduler=scheduler, hparams=hparams)

        for t in range(50):
            y_pred = self.model(self.x)
            loss = self.loss_fn(y_pred, self.y)
            loss.backward()
            train_op()

    def test_BertAdam(self):
        r"""Tests BertAdam.
        """
        optimizer = BertAdam(self.model.parameters())

        for t in range(50):
            y_pred = self.model(self.x)
            loss = self.loss_fn(y_pred, self.y)
            loss.backward()
            optimizer.step()


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various data iterator utility classes.
"""

import pkg_resources

from torch import __version__ as _torch_version  # type: ignore
from torch.utils.data.dataloader import (  # type: ignore
    _BaseDataLoaderIter, _SingleProcessDataLoaderIter,
    _MultiProcessingDataLoaderIter)


_torch_version = pkg_resources.parse_version(_torch_version)


# PyTorch 1.3 change some attribute names in `_BaseDataLoaderIter`,
# `_SingleProcessDataLoaderIter`, and `_MultiProcessingDataLoaderIter`

if _torch_version >= pkg_resources.parse_version("1.3.0"):

    class TexarBaseDataLoaderIter(_BaseDataLoaderIter):

        @property
        def dataset(self):
            return self._dataset

        @property
        def dataset_kind(self):
            return self._dataset_kind

        @property
        def auto_collation(self):
            return self._auto_collation

        @property
        def drop_last(self):
            return self._drop_last

        @property
        def index_sampler(self):
            return self._index_sampler

        @property
        def num_workers(self):
            return self._num_workers

        @property
        def pin_memory(self):
            return self._pin_memory

        @property
        def timeout(self):
            return self._timeout

        @property
        def collate_fn(self):
            return self._collate_fn

        @property
        def sampler_iter(self):
            return self._sampler_iter

        @property
        def base_seed(self):
            return self._base_seed

    class TexarSingleProcessDataLoaderIter(_SingleProcessDataLoaderIter,
                                           TexarBaseDataLoaderIter):

        @property
        def dataset_fetcher(self):
            return self._dataset_fetcher

    class TexarMultiProcessingDataLoaderIter(_MultiProcessingDataLoaderIter,
                                             TexarBaseDataLoaderIter):

        @property
        def worker_init_fn(self):
            return self._worker_init_fn

        @property
        def worker_queue_idx_cycle(self):
            return self._worker_queue_idx_cycle

        @property
        def worker_result_queue(self):
            return self._worker_result_queue

        @property
        def worker_pids_set(self):
            return self._worker_pids_set

        @property
        def shutdown(self):
            return self._shutdown

        @property
        def send_idx(self):
            return self._send_idx

        @send_idx.setter
        def send_idx(self, value):
            self._send_idx = value

        @property
        def rcvd_idx(self):
            return self._rcvd_idx

        @property
        def task_info(self):
            return self._task_info

        @property
        def tasks_outstanding(self):
            return self._tasks_outstanding

        @tasks_outstanding.setter
        def tasks_outstanding(self, value):
            self._tasks_outstanding = value

        @property
        def workers_done_event(self):
            return self._workers_done_event

        @property
        def index_queues(self):
            return self._index_queues

        @property
        def workers(self):
            return self._workers

        @property
        def workers_status(self):
            return self._workers_status

        @property
        def data_queue(self):
            return self._data_queue

        @property
        def pin_memory_thread_done_event(self):
            if hasattr(self, '_pin_memory_thread_done_event'):
                return self._pin_memory_thread_done_event

        @property
        def pin_memory_thread(self):
            if hasattr(self, '_pin_memory_thread'):
                return self._pin_memory_thread

else:
    class TexarBaseDataLoaderIter(_BaseDataLoaderIter):  # type: ignore
        pass

    class TexarSingleProcessDataLoaderIter(  # type: ignore
        _SingleProcessDataLoaderIter, TexarBaseDataLoaderIter):
        pass

    class TexarMultiProcessingDataLoaderIter(  # type: ignore
        _MultiProcessingDataLoaderIter, TexarBaseDataLoaderIter):
        pass



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Regularizers
"""

# pylint: disable=redefined-outer-name

from typing import Dict, Union

import torch

__all__ = [
    'Regularizer',
    'L1L2',
    'l1',
    'l2',
    'l1_l2',
]


class Regularizer:
    r"""Regularizer base class.
    """

    def __call__(self, x):
        return 0.

    @classmethod
    def from_config(cls, config):
        r"""Construct a :class:`Regularizer` instance given configurations.
        """
        return cls(**config)

    def get_config(self) -> Dict[str, float]:
        r"""Return a Dict with configurations for the current regularizer
        instance.
        """
        raise NotImplementedError


class L1L2(Regularizer):
    r"""Regularizer for L1 and L2 regularization.

    Args:
        l1: Float or Int; L1 regularization factor.
        l2: Float or Int; L2 regularization factor.
    """

    def __init__(self, l1: Union[int, float] = 0., l2: Union[int, float] = 0.):
        self.l1 = float(l1)
        self.l2 = float(l2)

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        regularization = torch.tensor(0.)
        if self.l1:
            regularization += torch.sum(self.l1 * torch.abs(x).float())
        if self.l2:
            regularization += torch.sum(self.l2 * (x ** 2).float())
        return regularization

    def get_config(self) -> Dict[str, float]:
        return {'l1': float(self.l1), 'l2': float(self.l2)}


def l1(l: Union[int, float] = 0.01) -> Regularizer:
    r"""
    Construct an L1 regularizer.

    Args:
        l: Float or Int
            L1 regularization factor.
    Returns:
        An L1L2 regularization instance with l1=l
    """
    return L1L2(l1=l)


def l2(l: Union[int, float] = 0.01) -> Regularizer:
    r"""
    Construct an L2 regularizer.

    Args:
        l: Float or Int
            L2 regularization factor.
    Returns:
        An L1L2 regularization instance with l2=l
    """
    return L1L2(l2=l)


def l1_l2(l1: Union[int, float] = 0.01,
          l2: Union[int, float] = 0.01) -> Regularizer:
    r"""
    Construct a regularizer with both L1 and L2 components.

    Args:
        l1: Float or Int
            L1 regularization factor.
        l2: Float or Int
            L2 regularization factor.
    Returns:
        An L1L2 regularization instance with above regularization factors
    """
    return L1L2(l1=l1, l2=l2)



"""
Various utilities for data module
"""

from enum import Enum
from typing import (
    Any, Dict, ItemsView, KeysView, List, Optional, Tuple, Union, ValuesView)

import numpy as np

__all__ = [
    'padded_batch',
    'connect_name',
    'Batch',
    '_LazyStrategy',
    '_CacheStrategy',
]


def padded_batch(examples: Union[List[np.ndarray], List[List[int]]],
                 pad_length: Optional[int] = None, pad_value: int = 0) \
        -> Tuple[np.ndarray, List[int]]:
    r"""Pad a batch of integer lists (or numpy arrays) to the same length, and
    stack them together.

    Args:
        examples (list of lists): The list of examples.
        pad_length (int, optional): The desired length after padding. If
            `None`, use the maximum length of lists in the batch. Defaults to
            `None`. Note that if ``pad_length`` is not `None` and the
            maximum length of lists is greater than ``pad_length``, then all
            lists are padded to the maximum length instead.
        pad_value (int, optional): The value to fill in the padded positions.
            Defaults to 0.

    Returns:
        A tuple of two elements, with the first being the padded batch, and the
        second being the original lengths of each list.
    """
    lengths = [len(sent) for sent in examples]
    pad_length = pad_length or max(lengths)

    padded = np.full((len(examples), pad_length), pad_value, dtype=np.int64)
    for b_idx, sent in enumerate(examples):
        length = lengths[b_idx]
        padded[b_idx, :length] = sent[:length]
    return padded, lengths


def connect_name(lhs_name, rhs_name):
    if not lhs_name:
        return rhs_name
    if not rhs_name:
        return lhs_name
    return "{}_{}".format(lhs_name, rhs_name)


class Batch:
    r"""Wrapper over Python dictionaries representing a batch. It provides a
    dictionary-like interface to access its fields. This class can be used in
    the followed way

    .. code-block:: python

        hparams = {
            'dataset': { 'files': 'data.txt', 'vocab_file': 'vocab.txt' },
            'batch_size': 1
        }

        data = MonoTextData(hparams)
        iterator = DataIterator(data)

        for batch in iterator:
            # batch is Batch object and contains the following fields
            # batch == {
            #    'text': [['<BOS>', 'example', 'sequence', '<EOS>']],
            #    'text_ids': [[1, 5, 10, 2]],
            #    'length': [4]
            # }

            input_ids = torch.tensor(batch['text_ids'])

            # we can also access the elements using dot notation
            input_text = batch.text
    """

    def __init__(self, batch_size: int, batch: Optional[Dict[str, Any]] = None,
                 **kwargs):
        self.batch_size = batch_size
        self._batch = batch or {}
        if isinstance(self._batch, dict):
            self._batch.update(kwargs)

    def __getattr__(self, item):
        if item not in super().__getattribute__('_batch'):
            raise AttributeError
        return self._batch[item]

    def __getitem__(self, item):
        return self._batch[item]

    def __len__(self) -> int:
        return self.batch_size

    def __repr__(self):
        return repr(self._batch)

    def keys(self) -> KeysView[str]:
        return self._batch.keys()

    def values(self) -> ValuesView[Any]:
        return self._batch.values()

    def items(self) -> ItemsView[str, Any]:
        return self._batch.items()


class _LazyStrategy(Enum):
    NONE = "none"
    PROCESS = "process"
    ALL = "all"


class _CacheStrategy(Enum):
    NONE = "none"
    LOADED = "loaded"
    PROCESSED = "processed"



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various RNN encoders.
"""

from typing import Any, Dict, Generic, List, Optional, Tuple, TypeVar, Union

import torch
import torch.nn as nn

from texar.torch.core import layers
from texar.torch.core.cell_wrappers import LSTMCell, RNNCellBase
from texar.torch.hyperparams import HParams
from texar.torch.modules.encoders.encoder_base import EncoderBase
from texar.torch.modules.networks.conv_networks import _to_list
from texar.torch.utils.rnn import bidirectional_dynamic_rnn, dynamic_rnn
from texar.torch.utils.shapes import mask_sequences

__all__ = [
    "_forward_output_layers",
    "RNNEncoderBase",
    "UnidirectionalRNNEncoder",
    "BidirectionalRNNEncoder",
]

State = TypeVar('State')


def _default_output_layer_hparams() -> Dict[str, Any]:
    return {
        "num_layers": 0,
        "layer_size": 128,
        "activation": "Identity",
        "final_layer_activation": None,
        "other_dense_kwargs": None,
        "dropout_layer_ids": [],
        "dropout_rate": 0.5,
        "variational_dropout": False,
        "@no_typecheck": ["activation", "final_layer_activation",
                          "layer_size", "dropout_layer_ids"]
    }


def _build_dense_output_layer(cell_output_size: int,
                              hparams: HParams) -> Optional[nn.Sequential]:
    r"""Build the output layers.

    Args:
        cell_output_size: The output size of the rnn cell.
        hparams (dict or HParams): Hyperparameters. Missing hyperparameters
            will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure and
            default values.

    Returns:
        A :torch_nn:`Sequential` module containing the output layers.
    """
    nlayers = hparams.num_layers

    if nlayers <= 0:
        return None

    layer_size = _to_list(
        hparams.layer_size, 'output_layer.layer_size', nlayers)

    dropout_layer_ids = _to_list(hparams.dropout_layer_ids)

    other_kwargs = hparams.other_dense_kwargs or {}
    if isinstance(other_kwargs, HParams):
        other_kwargs = other_kwargs.todict()
    if not isinstance(other_kwargs, dict):
        raise ValueError(
            "hparams 'output_layer.other_dense_kwargs' must be a dict.")

    output_layers: List[nn.Module] = []
    for i in range(nlayers):
        if i in dropout_layer_ids:
            # TODO: Variational dropout is not implemented.
            output_layers.append(nn.Dropout(p=hparams.dropout_rate))

        dense_layer = nn.Linear(in_features=(cell_output_size if i == 0
                                             else layer_size[i - 1]),
                                out_features=layer_size[i], **other_kwargs)

        output_layers.append(dense_layer)

        if i == nlayers - 1:
            activation = hparams.final_layer_activation
        else:
            activation = hparams.activation

        if activation is not None:
            layer_hparams = {"type": activation, "kwargs": {}}
            activation_layer = layers.get_layer(hparams=layer_hparams)
            output_layers.append(activation_layer)

    if nlayers in dropout_layer_ids:
        output_layers.append(nn.Dropout(p=hparams.dropout_rate))

    return nn.Sequential(*output_layers)


def _forward_output_layers(
        inputs: torch.Tensor,
        output_layer: Optional[nn.Module],
        time_major: bool,
        sequence_length: Optional[Union[torch.LongTensor, List[int]]] = None) \
        -> Tuple[torch.Tensor, int]:
    r"""Forwards inputs through the output layers.

    Args:
        inputs: A Tensor of shape ``[batch_size, max_time] + input_size`` if
            :attr:`time_major` is `False`, or shape
            ``[max_time, batch_size] + input_size`` if :attr:`time_major` is
            `True`.
        output_layer (optional): :torch_nn:`Sequential` or :torch_nn:`Module`
            of output layers.
        time_major (bool): The shape format of the :attr:`inputs` and
            :attr:`outputs` Tensors. If `True`, these tensors are of shape
            `[max_time, batch_size, input_size]`. If `False` (default),
            these tensors are of shape `[batch_size, max_time, input_size]`.
        sequence_length (optional): A 1D :tensor:`LongTensor` of shape
            ``[batch_size]``. Sequence lengths of the batch inputs. Used to
            copy-through state and zero-out outputs when past a batch element's
            sequence length.

    Returns:
        A pair :attr:`(outputs, outputs_size), where

        - :attr:`outputs`: A Tensor of shape
        `[batch_size, max_time] + outputs_size`.

        - :attr:`outputs_size`: An `int` representing the output size.
    """
    if output_layer is None:
        return inputs, inputs.shape[-1]

    output = output_layer(inputs)

    if sequence_length is not None:
        output = mask_sequences(output, sequence_length, time_major=time_major)

    output_size = output.shape[-1]

    return output, output_size


class RNNEncoderBase(EncoderBase, Generic[State]):
    r"""Base class for all RNN encoder classes to inherit.

    Args:
        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparameters will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure and
            default values.
    """

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "name": "rnn_encoder"
            }
        """
        return {
            'name': 'rnn_encoder'
        }


class UnidirectionalRNNEncoder(RNNEncoderBase[State]):
    r"""One directional RNN encoder.

    Args:
        cell: (RNNCell, optional) If not specified,
            a cell is created as specified in :attr:`hparams["rnn_cell"]`.
        output_layer (optional): An instance of
            :torch_nn:`Module`. Applies to the RNN cell
            output of each step. If `None` (default), the output layer is
            created as specified in :attr:`hparams["output_layer"]`.
        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparameters will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure and
            default values.

    See :meth:`forward` for the inputs and outputs of the encoder.

    Example:

    .. code-block:: python

        # Use with embedder
        embedder = WordEmbedder(vocab_size, hparams=emb_hparams)
        encoder = UnidirectionalRNNEncoder(hparams=enc_hparams)

        outputs, final_state = encoder(
            inputs=embedder(data_batch['text_ids']),
            sequence_length=data_batch['length'])

    .. document private functions
    """
    _cell: RNNCellBase[State]

    def __init__(self,
                 input_size: int,
                 cell: Optional[RNNCellBase[State]] = None,
                 output_layer: Optional[nn.Module] = None,
                 hparams=None):
        super().__init__(hparams=hparams)

        # Make RNN cell
        if cell is not None:
            self._cell = cell
        else:
            self._cell = layers.get_rnn_cell(input_size,
                                             self._hparams.rnn_cell)

        # Make output layer
        self._output_layer: Optional[nn.Module]
        if output_layer is not None:
            self._output_layer = output_layer
            self._output_layer_hparams = None
        else:
            self._output_layer = _build_dense_output_layer(
                self._cell.hidden_size, self._hparams.output_layer)
            self._output_layer_hparams = self._hparams.output_layer

    @staticmethod
    def default_hparams() -> Dict[str, Any]:
        r"""Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "rnn_cell": default_rnn_cell_hparams(),
                "output_layer": {
                    "num_layers": 0,
                    "layer_size": 128,
                    "activation": "identity",
                    "final_layer_activation": None,
                    "other_dense_kwargs": None,
                    "dropout_layer_ids": [],
                    "dropout_rate": 0.5,
                    "variational_dropout": False
                },
                "name": "unidirectional_rnn_encoder"
            }

        Here:

        `"rnn_cell"`: dict
            A dictionary of RNN cell hyperparameters. Ignored if
            :attr:`cell` is given to the encoder constructor.

            The default value is defined in
            :func:`~texar.torch.core.default_rnn_cell_hparams`.

        `"output_layer"`: dict
            Output layer hyperparameters. Ignored if :attr:`output_layer`
            is given to the encoder constructor. Includes:

            `"num_layers"`: int
                The number of output (dense) layers. Set to 0 to avoid any
                output layers applied to the cell outputs.

            `"layer_size"`: int or list
                The size of each of the output (dense) layers.

                If an `int`, each output layer will have the same size. If
                a list, the length must equal to :attr:`num_layers`.

            `"activation"`: str or callable or None
                Activation function for each of the output (dense)
                layer except for the final layer. This can be
                a function, or its string name or module path.
                If function name is given, the function must be from
                :mod:`torch.nn`.
                For example:

                .. code-block:: python

                    "activation": "relu" # function name
                    "activation": "my_module.my_activation_fn" # module path
                    "activation": my_module.my_activation_fn # function

                Default is `None` which results in an identity activation.

            `"final_layer_activation"`: str or callable or None
                The activation function for the final output layer.

            `"other_dense_kwargs"`: dict or None
                Other keyword arguments to construct each of the output
                dense layers, e.g., ``bias``. See
                :torch_nn:`Linear` for the keyword arguments.

            `"dropout_layer_ids"`: int or list
                The indexes of layers (starting from 0) whose inputs
                are applied with dropout. The index = :attr:`num_layers`
                means dropout applies to the final layer output. For example,

                .. code-block:: python

                    {
                        "num_layers": 2,
                        "dropout_layer_ids": [0, 2]
                    }

                will leads to a series of layers as
                `-dropout-layer0-layer1-dropout-`.

                The dropout mode (training or not) is controlled
                by :attr:`self.training`.

            `"dropout_rate"`: float
                The dropout rate, between 0 and 1. For example,
                ``"dropout_rate": 0.1`` would zero out 10% of elements.

            `"variational_dropout"`: bool
                Whether the dropout mask is the same across all time steps.

        `"name"`: str
            Name of the encoder
        """
        hparams = RNNEncoderBase.default_hparams()
        hparams.update({
            "rnn_cell": layers.default_rnn_cell_hparams(),
            "output_layer": _default_output_layer_hparams(),
            "name": "unidirectional_rnn_encoder"
        })
        return hparams

    def forward(self,  # type: ignore
                inputs: torch.Tensor,
                sequence_length: Optional[Union[torch.LongTensor,
                                                List[int]]] = None,
                initial_state: Optional[State] = None,
                time_major: bool = False,
                return_cell_output: bool = False,
                return_output_size: bool = False):
        r"""Encodes the inputs.

        Args:
            inputs: A 3D Tensor of shape ``[batch_size, max_time, dim]``.
                The first two dimensions
                :attr:`batch_size` and :attr:`max_time` are exchanged if
                :attr:`time_major` is `True`.
            sequence_length (optional): A 1D :tensor:`LongTensor` of shape
                ``[batch_size]``.
                Sequence lengths of the batch inputs. Used to copy-through
                state and zero-out outputs when past a batch element's sequence
                length.
            initial_state (optional): Initial state of the RNN.
            time_major (bool): The shape format of the :attr:`inputs` and
                :attr:`outputs` Tensors. If `True`, these tensors are of shape
                ``[max_time, batch_size, depth]``. If `False` (default),
                these tensors are of shape ``[batch_size, max_time, depth]``.
            return_cell_output (bool): Whether to return the output of the RNN
                cell. This is the results prior to the output layer.
            return_output_size (bool): Whether to return the size of the
                output (i.e., the results after output layers).

        Returns:
            - By default (both ``return_cell_output`` and ``return_output_size``
              are `False`), returns a pair :attr:`(outputs, final_state)`,
              where

              - :attr:`outputs`: The RNN output tensor by the output layer
                (if exists) or the RNN cell (otherwise). The tensor is of
                shape ``[batch_size, max_time, output_size]`` if
                ``time_major`` is `False`, or
                ``[max_time, batch_size, output_size]`` if
                ``time_major`` is `True`.
                If RNN cell output is a (nested) tuple of Tensors, then the
                :attr:`outputs` will be a (nested) tuple having the same
                nest structure as the cell output.

              - :attr:`final_state`: The final state of the RNN, which is a
                Tensor of shape ``[batch_size] + cell.state_size`` or
                a (nested) tuple of Tensors if ``cell.state_size`` is a
                (nested) tuple.

            - If ``return_cell_output`` is True, returns a triple
              :attr:`(outputs, final_state, cell_outputs)`

              - :attr:`cell_outputs`: The outputs by the RNN cell prior to the
                output layer, having the same structure with :attr:`outputs`
                except for the ``output_dim``.

            - If ``return_output_size`` is `True`, returns a tuple
              :attr:`(outputs, final_state, output_size)`

              - :attr:`output_size`: A (possibly nested tuple of) int
                representing the size of :attr:`outputs`. If a single int or
                an int array, then ``outputs`` has shape
                ``[batch/time, time/batch] + output_size``. If
                a (nested) tuple, then ``output_size`` has the same
                structure as with ``outputs``.

            - If both ``return_cell_output`` and ``return_output_size`` are
              `True`, returns
              :attr:`(outputs, final_state, cell_outputs, output_size)`.
        """

        cell_outputs, state = dynamic_rnn(
            cell=self._cell,
            inputs=inputs,
            sequence_length=sequence_length,
            initial_state=initial_state,
            time_major=time_major)

        outputs, output_size = _forward_output_layers(
            inputs=cell_outputs,
            output_layer=self._output_layer,
            time_major=time_major,
            sequence_length=sequence_length)

        rets = (outputs, state)
        if return_cell_output:
            rets += (cell_outputs,)  # type: ignore
        if return_output_size:
            rets += (output_size,)  # type: ignore
        return rets

    @property
    def cell(self) -> RNNCellBase[State]:
        r"""The RNN cell.
        """
        return self._cell

    @property
    def state_size(self) -> int:
        r"""The state size of encoder cell.
        Same as :attr:`encoder.cell.state_size`.
        """
        if isinstance(self._cell, LSTMCell):
            return 2 * self._cell.hidden_size  # type: ignore
        else:
            return self._cell.hidden_size

    @property
    def output_layer(self) -> Optional[nn.Module]:
        r"""The output layer.
        """
        return self._output_layer

    @property
    def output_size(self) -> int:
        r"""The feature size of :meth:`forward` output :attr:`outputs`.
        If output layer does not exist, the feature size is equal to
        :attr:`encoder.cell.hidden_size`, otherwise the feature size
        is equal to last dimension value of output layer output size.
        """
        # TODO: We will change the implementation to
        # something that does not require a forward pass.

        dim = self._cell.hidden_size
        if self._output_layer is not None:
            dummy_tensor = torch.Tensor(dim)
            dim = self._output_layer(dummy_tensor).size(-1)
        return dim


class BidirectionalRNNEncoder(RNNEncoderBase):
    r"""Bidirectional forward-backward RNN encoder.

    Args:
        cell_fw (RNNCell, optional): The forward RNN cell. If not given,
            a cell is created as specified in ``hparams["rnn_cell_fw"]``.
        cell_bw (RNNCell, optional): The backward RNN cell. If not given,
            a cell is created as specified in ``hparams["rnn_cell_bw"]``.
        output_layer_fw (optional): An instance of
            :torch_nn:`Module`. Apply to the forward
            RNN cell output of each step. If `None` (default), the output
            layer is created as specified in ``hparams["output_layer_fw"]``.
        output_layer_bw (optional): An instance of
            :torch_nn:`Module`. Apply to the backward
            RNN cell output of each step. If `None` (default), the output
            layer is created as specified in ``hparams["output_layer_bw"]``.
        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparameters will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure and
            default values.

    See :meth:`forward` for the inputs and outputs of the encoder.

    Example:

        .. code-block:: python

            # Use with embedder
            embedder = WordEmbedder(vocab_size, hparams=emb_hparams)
            encoder = BidirectionalRNNEncoder(hparams=enc_hparams)

            outputs, final_state = encoder(
                inputs=embedder(data_batch['text_ids']),
                sequence_length=data_batch['length'])
            # outputs == (outputs_fw, outputs_bw)
            # final_state == (final_state_fw, final_state_bw)

    .. document private functions
    """

    def __init__(self,
                 input_size: int,
                 cell_fw: Optional[RNNCellBase[State]] = None,
                 cell_bw: Optional[RNNCellBase[State]] = None,
                 output_layer_fw: Optional[nn.Module] = None,
                 output_layer_bw: Optional[nn.Module] = None,
                 hparams=None):
        super().__init__(hparams=hparams)

        # Make RNN cells
        if cell_fw is not None:
            self._cell_fw = cell_fw
        else:
            self._cell_fw = layers.get_rnn_cell(input_size,
                                                self._hparams.rnn_cell_fw)

        if cell_bw is not None:
            self._cell_bw = cell_bw
        elif self._hparams.rnn_cell_share_config:
            self._cell_bw = layers.get_rnn_cell(input_size,
                                                self._hparams.rnn_cell_fw)
        else:
            self._cell_bw = layers.get_rnn_cell(input_size,
                                                self._hparams.rnn_cell_bw)

        # Make output layers

        self.__output_layer_fw: Optional[nn.Module]
        if output_layer_fw is not None:
            self._output_layer_fw = output_layer_fw
            self._output_layer_hparams_fw = None
        else:
            self._output_layer_fw = _build_dense_output_layer(  # type: ignore
                self._cell_fw.hidden_size, self._hparams.output_layer_fw)
            self._output_layer_hparams_fw = self._hparams.output_layer_fw

        self.__output_layer_bw: Optional[nn.Module]
        if output_layer_bw is not None:
            self._output_layer_bw = output_layer_bw
            self._output_layer_hparams_bw = None
        elif self._hparams.output_layer_share_config:
            self._output_layer_bw = _build_dense_output_layer(  # type: ignore
                self._cell_bw.hidden_size, self._hparams.output_layer_fw)
            self._output_layer_hparams_bw = self._hparams.output_layer_fw
        else:
            self._output_layer_bw = _build_dense_output_layer(  # type: ignore
                self._cell_bw.hidden_size, self._hparams.output_layer_bw)
            self._output_layer_hparams_bw = self._hparams.output_layer_bw

    @staticmethod
    def default_hparams() -> Dict[str, Any]:
        r"""Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "rnn_cell_fw": default_rnn_cell_hparams(),
                "rnn_cell_bw": default_rnn_cell_hparams(),
                "rnn_cell_share_config": True,
                "output_layer_fw": {
                    "num_layers": 0,
                    "layer_size": 128,
                    "activation": "identity",
                    "final_layer_activation": None,
                    "other_dense_kwargs": None,
                    "dropout_layer_ids": [],
                    "dropout_rate": 0.5,
                    "variational_dropout": False
                },
                "output_layer_bw": {
                    # Same hyperparams and default values as "output_layer_fw"
                    # ...
                },
                "output_layer_share_config": True,
                "name": "bidirectional_rnn_encoder"
            }

        Here:

        `"rnn_cell_fw"`: dict
            Hyperparameters of the forward RNN cell.
            Ignored if :attr:`cell_fw` is given to the encoder constructor.

            The default value is defined in
            :func:`~texar.torch.core.default_rnn_cell_hparams`.

        `"rnn_cell_bw"`: dict
            Hyperparameters of the backward RNN cell.
            Ignored if :attr:`cell_bw` is given to the encoder constructor,
            or if `"rnn_cell_share_config"` is `True`.

            The default value is defined in
            :meth:`~texar.torch.core.default_rnn_cell_hparams`.

        `"rnn_cell_share_config"`: bool
            Whether share hyperparameters of the backward cell with the
            forward cell. Note that the cell parameters (variables) are not
            shared.

        `"output_layer_fw"`: dict
            Hyperparameters of the forward output layer. Ignored if
            ``output_layer_fw`` is given to the constructor.
            See the ``"output_layer"`` field of
            :meth:`~texar.torch.modules.UnidirectionalRNNEncoder` for details.

        `"output_layer_bw"`: dict
            Hyperparameters of the backward output layer. Ignored if
            :attr:`output_layer_bw` is given to the constructor. Have the
            same structure and defaults with :attr:`"output_layer_fw"`.

            Ignored if ``output_layer_share_config`` is `True`.

        `"output_layer_share_config"`: bool
            Whether share hyperparameters of the backward output layer
            with the forward output layer. Note that the layer parameters
            (variables) are not shared.

        `"name"`: str
            Name of the encoder
        """
        hparams = RNNEncoderBase.default_hparams()
        hparams.update({
            "rnn_cell_fw": layers.default_rnn_cell_hparams(),
            "rnn_cell_bw": layers.default_rnn_cell_hparams(),
            "rnn_cell_share_config": True,
            "output_layer_fw": _default_output_layer_hparams(),
            "output_layer_bw": _default_output_layer_hparams(),
            "output_layer_share_config": True,
            "name": "bidirectional_rnn_encoder"
        })
        return hparams

    def forward(self,  # type: ignore
                inputs: torch.Tensor,
                sequence_length: Optional[Union[torch.LongTensor,
                                                List[int]]] = None,
                initial_state_fw: Optional[State] = None,
                initial_state_bw: Optional[State] = None,
                time_major: bool = False,
                return_cell_output: bool = False,
                return_output_size: bool = False):
        r"""Encodes the inputs.

        Args:
            inputs: A 3D Tensor of shape ``[batch_size, max_time, dim]``.
                The first two dimensions
                ``batch_size`` and ``max_time`` may be exchanged if
                ``time_major`` is `True`.
            sequence_length (optional): A 1D :tensor:`LongTensor` of shape
                ``[batch_size]``.
                Sequence lengths of the batch inputs. Used to copy-through
                state and zero-out outputs when past a batch element's sequence
                length.
            initial_state_fw: (optional): Initial state of the forward RNN.
            initial_state_bw: (optional): Initial state of the backward RNN.
            time_major (bool): The shape format of the :attr:`inputs` and
                :attr:`outputs` Tensors. If `True`, these tensors are of shape
                ``[max_time, batch_size, depth]``. If `False` (default),
                these tensors are of shape ``[batch_size, max_time, depth]``.
            return_cell_output (bool): Whether to return the output of the RNN
                cell. This is the results prior to the output layer.
            return_output_size (bool): Whether to return the output size of the
                RNN cell. This is the results after the output layer.

        Returns:
            - By default (both ``return_cell_output`` and ``return_output_size``
              are `False`), returns a pair :attr:`(outputs, final_state)`

              - :attr:`outputs`: A tuple ``(outputs_fw, outputs_bw)``
                containing the forward and the backward RNN outputs, each of
                which is of shape ``[batch_size, max_time, output_dim]``
                if ``time_major`` is `False`, or
                ``[max_time, batch_size, output_dim]`` if ``time_major``
                is `True`.
                If RNN cell output is a (nested) tuple of Tensors, then
                ``outputs_fw`` and ``outputs_bw`` will be a (nested) tuple
                having the same structure as the cell output.

              - :attr:`final_state`: A tuple
                ``(final_state_fw, final_state_bw)`` containing the final
                states of the forward and backward RNNs, each of which is a
                Tensor of shape ``[batch_size] + cell.state_size``, or a
                (nested) tuple of Tensors if ``cell.state_size`` is a (nested)
                tuple.

            - If ``return_cell_output`` is `True`, returns a triple
              :attr:`(outputs, final_state, cell_outputs)` where

              - :attr:`cell_outputs`: A tuple
                ``(cell_outputs_fw, cell_outputs_bw)`` containing the outputs
                by the forward and backward RNN cells prior to the output
                layers, having the same structure with :attr:`outputs` except
                for the ``output_dim``.

            - If ``return_output_size`` is `True`, returns a tuple
              :attr:`(outputs, final_state, output_size)` where

              - :attr:`output_size`: A tuple
                ``(output_size_fw, output_size_bw)`` containing the size of
                ``outputs_fw`` and ``outputs_bw``, respectively.
                Take ``*_fw`` for example, ``output_size_fw`` is a (possibly
                nested tuple of) int. If a single int or an int array, then
                ``outputs_fw`` has shape
                ``[batch/time, time/batch] + output_size_fw``. If a (nested)
                tuple, then ``output_size_fw`` has the same structure as
                ``outputs_fw``. The same applies to ``output_size_bw``.

            - If both ``return_cell_output`` and ``return_output_size`` are
              `True`, returns
              :attr:`(outputs, final_state, cell_outputs, output_size)`.
        """

        cell_outputs, states = bidirectional_dynamic_rnn(
            cell_fw=self._cell_fw,
            cell_bw=self._cell_bw,
            inputs=inputs,
            sequence_length=sequence_length,
            initial_state_fw=initial_state_fw,
            initial_state_bw=initial_state_bw,
            time_major=time_major)

        outputs_fw, output_size_fw = _forward_output_layers(
            inputs=cell_outputs[0],
            output_layer=self._output_layer_fw,
            time_major=time_major,
            sequence_length=sequence_length)

        outputs_bw, output_size_bw = _forward_output_layers(
            inputs=cell_outputs[1],
            output_layer=self._output_layer_bw,
            time_major=time_major,
            sequence_length=sequence_length)

        outputs = (outputs_fw, outputs_bw)
        output_size = (output_size_fw, output_size_bw)

        returns = (outputs, states)
        if return_cell_output:
            returns += (cell_outputs,)  # type: ignore
        if return_output_size:
            returns += (output_size,)  # type: ignore
        return returns

    @property
    def cell_fw(self) -> RNNCellBase[State]:
        r"""The forward RNN cell.
        """
        return self._cell_fw

    @property
    def cell_bw(self) -> RNNCellBase[State]:
        r"""The backward RNN cell.
        """
        return self._cell_bw

    @property
    def state_size_fw(self) -> int:
        r"""The state size of the forward encoder cell.
        Same as :attr:`encoder.cell_fw.state_size`.
        """
        if isinstance(self._cell_fw, LSTMCell):
            return 2 * self._cell_fw.hidden_size  # type: ignore
        else:
            return self._cell_fw.hidden_size

    @property
    def state_size_bw(self) -> int:
        r"""The state size of the backward encoder cell.
        Same as :attr:`encoder.cell_bw.state_size`.
        """
        if isinstance(self._cell_bw, LSTMCell):
            return 2 * self._cell_bw.hidden_size  # type: ignore
        else:
            return self._cell_bw.hidden_size

    @property
    def output_layer_fw(self) -> Optional[nn.Module]:
        r"""The output layer of the forward RNN.
        """
        return self._output_layer_fw

    @property
    def output_layer_bw(self) -> Optional[nn.Module]:
        r"""The output layer of the backward RNN.
        """
        return self._output_layer_bw

    @property
    def output_size(self) -> Tuple[int, int]:
        r"""The feature sizes of :meth:`forward` outputs
        :attr:`output_size_fw` and :attr:`output_size_bw`.
        Each feature size is equal to last dimension
        value of corresponding result size.
        """
        # TODO: We will change the implementation to
        # something that does not require a forward pass.
        dim_bw = self._cell_bw.hidden_size
        dim_fw = self._cell_fw.hidden_size
        if self._output_layer_bw is not None:
            dummy_tensor_bw = torch.Tensor(dim_bw)
            output_bw = self._output_layer_bw(dummy_tensor_bw).size()[-1]
        else:
            output_bw = dim_bw
        if self._output_layer_fw is not None:
            dummy_tensor_fw = torch.Tensor(dim_fw)
            output_fw = self._output_layer_fw(dummy_tensor_fw).size()[-1]
        else:
            output_fw = dim_fw
        return (output_fw, output_bw)



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Data class that supports reading pickled data as record structures.
"""
import copy
import io
import pickle
import warnings
from enum import Enum
from typing import (
    Any, Callable, Dict, List, NamedTuple, Optional, Tuple, TypeVar, Union)

import numpy as np
import torch

from texar.torch.data.data.data_base import DatasetBase, DataSource
from texar.torch.data.data.dataset_utils import Batch, padded_batch
from texar.torch.hyperparams import HParams
from texar.torch.utils.dtypes import get_numpy_dtype
from texar.torch.utils.types import MaybeList

__all__ = [
    "_default_record_dataset_hparams",
    "PickleDataSource",
    "RecordData",
]


def _default_record_dataset_hparams():
    r"""Returns hyperparameters of a record dataset with default values.

    See :meth:`texar.torch.data.RecordData.default_hparams` for details.
    """
    return {
        "files": [],
        "feature_types": None,
        "feature_original_types": None,
        "feature_convert_types": {},
        "image_options": {},
        "compression_type": None,
        "other_transformations": [],
        "num_shards": None,
        "shard_id": None,
        "data_name": None,
        "@no_typecheck": [
            "files",
            "feature_types",
            "feature_original_types",
            "feature_convert_types",
            "image_options"
        ],
    }


RawExample = TypeVar('RawExample')


class PickleDataSource(DataSource[RawExample]):
    r"""Data source for reading from (multiple) pickled binary files. Each file
    could contain multiple pickled objects, and each object is yielded as an
    example.

    This data source does not support indexing.

    Args:
        file_paths (str or list[str]): Paths to pickled binary files.
        lists_are_examples (bool): If `True`, lists will be treated as
            a single example; if `False`, each element in the list will be
            treated as separate examples. Default is `True`. Set this to
            `False` if the entire pickled binary file is a list.

            .. note::
                It is recommended against storing all examples as a list,
                because in this case, all examples can only be accessed
                after the whole list is parsed.

        pickle_kwargs: Additional keyword arguments to pass to
            :meth:`pickle.load`.
    """

    def __init__(self, file_paths: MaybeList[str],
                 lists_are_examples: bool = True, **pickle_kwargs):
        if isinstance(file_paths, str):
            file_paths = [file_paths]
        self._file_paths = file_paths
        self._lists_are_examples = lists_are_examples
        self._pickle_kwargs = pickle_kwargs

    def __iter__(self):
        for path in self._file_paths:
            with open(path, 'rb') as f:
                if self._lists_are_examples:
                    while True:
                        try:
                            ex = pickle.load(f, **self._pickle_kwargs)
                            if isinstance(ex, list):
                                yield from ex
                            else:
                                yield ex
                        except EOFError:
                            break
                else:
                    while True:
                        try:
                            yield pickle.load(f, **self._pickle_kwargs)
                        except EOFError:
                            break


TransformFn = Callable[[bytes], torch.ByteTensor]


def _create_image_transform(height: Optional[int], width: Optional[int],
                            resize_method: Union[str, int] = 'bilinear') \
        -> TransformFn:
    r"""Create a function based on `Pillow image transforms
    <https://pillow.readthedocs.io/en/3.1.x/reference/Image.html#PIL.Image.Image.resize>`
    that performs resizing with desired resize method (interpolation).

    Args:
        height (int, optional): Height of the transformed image. Set to `None`
            to not perform resizing.
        width (int, optional): Width of the transformed image. Set to `None`
            to not perform resizing.
        resize_method (str or int, optional): Interpolation method to use.
            Supported values are ``"nearest"`` (nearest neighbor),
            ``"bilinear"``, ``"bicubic"``, and ``"lanczos"``. Enum values from
            PIL (e.g., ``PIL.Image.BILINEAR``) are also supported. Defaults to
            ``"bilinear"``.

    Returns:
        The created transformation function.
    """
    try:
        import PIL.Image
    except ImportError:
        raise ImportError(
            "To use image resizing with RecordData, the Pillow library must be "
            "installed. Please see "
            "https://pillow.readthedocs.io/en/stable/installation.html.")

    # We take the final part of a possibly dot-separated string for
    # compatibility reasons, because in Texar-TF `resize_method` could take the
    # form of "tf.image.ResizeMethod.BILINEAR".
    if isinstance(resize_method, int):
        interpolation = resize_method
    else:
        method = resize_method.lower().split('.')[-1]
        if method in ["nearest_neighbor", "nearest"]:
            interpolation = PIL.Image.NEAREST
        elif method == "bilinear":
            interpolation = PIL.Image.BILINEAR
        elif method == "bicubic":
            interpolation = PIL.Image.BICUBIC
        elif method == "lanczos":
            interpolation = PIL.Image.LANCZOS
        else:
            raise ValueError(f"Unsupported resize method '{resize_method}'")
    if height is None or width is None:
        size = None
    else:
        size = (height, width)

    def transform(raw_bytes):
        image = PIL.Image.open(io.BytesIO(raw_bytes))
        if size is not None:
            image = image.resize(size, interpolation)

        # Convert to torch Tensor. Adapted from
        # torchvision.transform.functional.to_tensor.
        if image.mode == '1':
            tensor = 255 * torch.from_numpy(
                np.array(image, np.uint8, copy=False))
        else:
            tensor = torch.ByteTensor(
                torch.ByteStorage.from_buffer(image.tobytes()))
        # PIL image mode: L, LA, P, I, F, RGB, YCbCr, RGBA, CMYK.
        if image.mode == 'YCbCr':
            n_channel = 3
        elif image.mode == 'I;16':
            n_channel = 1
        else:
            n_channel = len(image.mode)
        tensor = tensor.view(image.size[1], image.size[0], n_channel)
        return tensor

    return transform


class CollateMethod(Enum):
    StackedTensor = "stacked_tensor"
    PaddedTensor = "padded_tensor"
    List = "list"


class FeatureDescription(NamedTuple):
    r"""Description of a feature."""
    collate_method: CollateMethod
    dtype: Optional[np.dtype]
    shape: Optional[Tuple[int, ...]]


_DEPRECATED_NAME_MAPPING = {
    "FixedLenFeature": "stacked_tensor",
    "FixedLenSequenceFeature": "padded_tensor",
    "VarLenFeature": "list",
}


def _convert_feature_hparams(feature_types: Union[Dict[str, Any], HParams]) \
        -> Dict[str, FeatureDescription]:
    if isinstance(feature_types, HParams):
        feature_types = feature_types.todict()
    else:
        feature_types = copy.deepcopy(feature_types)
    show_deprecation_warning = False
    for key, value in feature_types.items():
        if len(value) > 1 and value[1] in _DEPRECATED_NAME_MAPPING:
            feature_types[key] = (
                value[0], _DEPRECATED_NAME_MAPPING[value[1]], *value[2:])
            show_deprecation_warning = True
    if show_deprecation_warning:
        warnings.warn(f"RecordData feature types "
                      f"{', '.join(repr(x) for x in _DEPRECATED_NAME_MAPPING)} "
                      f"are deprecated. Please see RecordData.default_hparams "
                      f"for update instructions.", UserWarning)

    features = {}
    for key, value in feature_types.items():
        shape: Optional[Tuple[int, ...]] = None
        if len(value) == 3:
            if isinstance(value[-1], int):
                shape = (value[-1],)
            elif all(isinstance(x, int) for x in value[-1]):
                shape = tuple(value[-1])
                if len(shape) == 0:
                    shape = (1,)  # scalar tensor
            else:
                raise ValueError(f"'shape' of feature '{key}' is not of type "
                                 f"int, tuple, or torch.Size")
        if len(value) < 2:
            collate_method = CollateMethod.StackedTensor
        else:
            try:
                collate_method = CollateMethod(value[1])
            except ValueError:
                values = [x.value for x in CollateMethod.__members__.values()]
                raise ValueError(
                    f"Unsupported feature collate method '{value[1]}' for "
                    f"feature '{key}', only "
                    f"{', '.join(repr(x) for x in values)} are "
                    f"supported as of now.")

        dtype = None
        if value[0] is not None:
            dtype = get_numpy_dtype(value[0])
        elif collate_method is not CollateMethod.List:
            raise ValueError(f"'dtype' for feature '{key}' must not be None "
                             f"unless collate method is 'list'")

        features[key] = FeatureDescription(collate_method, dtype, shape)

    return features


def _check_shape(tensor: np.ndarray, key: str, descriptor: FeatureDescription):
    if descriptor.shape is None:
        return
    # Check whether shape matches.
    if descriptor.collate_method is CollateMethod.PaddedTensor:
        shape = tensor.shape[1:]
    else:
        shape = tensor.shape
    if len(shape) == 0:
        shape = (1,)  # scalar tensor

    if shape != descriptor.shape:
        if descriptor.collate_method is CollateMethod.PaddedTensor:
            raise ValueError(
                f"Expected tensor of shape {('any', *descriptor.shape)} for "
                f"feature {key}, but received tensor of shape {tensor.shape}")
        else:
            raise ValueError(
                f"Expected tensor of shape {descriptor.shape} for "
                f"feature {key}, but received tensor of shape {tensor.shape}")


class RecordData(DatasetBase[Dict[str, Any], Dict[str, Any]]):
    r"""Record data which loads and processes pickled files.

    This module can be used to process image data, features, etc.

    Args:
        hparams (dict): Hyperparameters. See :meth:`default_hparams`
            for the defaults.
        device: The device of the produced batches. For GPU training, set to
            current CUDA device.

    The module reads and restores data from pickled files and results in a
    dataset whose element is a Python `dict` that maps feature names to feature
    values. The features names and dtypes are specified in
    :attr:`hparams.dataset.feature_types`.

    The module also provides simple processing options for image data, such
    as image resize.

    Example:

        .. code-block:: python

            # Read data from pickled file
            hparams={
                'dataset': {
                    'files': 'image1.pkl',
                    'feature_types': {
                        'height': ['int64', 'list'],  # or 'stacked_tensor'
                        'width': ['int64', 'list'],   # or 'stacked_tensor'
                        'label': ['int64', 'stacked_tensor'],
                        'image_raw': ['bytes', 'stacked_tensor'],
                    }
                },
                'batch_size': 1
            }
            data = RecordData(hparams)
            iterator = DataIterator(data)

            batch = next(iter(iterator))  # get the first batch in dataset
            # batch == {
            #    'data': {
            #        'height': [239],
            #        'width': [149],
            #        'label': tensor([1]),
            #
            #        # 'image_raw' is a NumPy ndarray of raw image bytes in this
            #        # example.
            #        'image_raw': [...],
            #    }
            # }

        .. code-block:: python

            # Read image data from pickled file and do resizing
            hparams={
                'dataset': {
                    'files': 'image2.pkl',
                    'feature_types': {
                        'label': ['int64', 'stacked_tensor'],
                        'image_raw': ['bytes', 'stacked_tensor'],
                    },
                    'image_options': {
                        'image_feature_name': 'image_raw',
                        'resize_height': 512,
                        'resize_width': 512,
                    }
                },
                'batch_size': 1
            }
            data = RecordData(hparams)
            iterator = DataIterator(data)

            batch = next(iter(iterator))  # get the first batch in dataset
            # batch == {
            #    'data': {
            #        'label': tensor([1]),
            #
            #        # "image_raw" is a tensor of image pixel data in this
            #        # example. Each image has a width of 512 and height of 512.
            #        'image_raw': tensor([...])
            #    }
            # }

    """

    def __init__(self, hparams=None, device: Optional[torch.device] = None,
                 data_source: Optional[DataSource] = None):
        self._hparams = HParams(hparams, self.default_hparams())

        feature_types = self._hparams.dataset.feature_original_types
        if feature_types is not None:
            warnings.warn(
                "'feature_original_types' of RecordData is deprecated. Please "
                "see default_hparams of RecordData for update instructions")
        if self._hparams.dataset.feature_types is not None:
            feature_types = self._hparams.dataset.feature_types
        elif feature_types is None:
            raise ValueError("'feature_types' must be specified")
        self._features = _convert_feature_hparams(feature_types)

        convert_types = self._hparams.dataset.feature_convert_types
        self._convert_types = {key: get_numpy_dtype(value)
                               for key, value in convert_types.items()}
        for key, dtype in self._convert_types.items():
            self._features[key] = self._features[key]._replace(dtype=dtype)

        image_options = self._hparams.dataset.image_options
        if isinstance(image_options, HParams):
            image_options = [image_options]
        self._image_transforms: Dict[str, TransformFn] = {}
        for options in image_options:
            key = options.get('image_feature_name')
            if key is None or key not in self._features:
                continue
            self._image_transforms[key] = _create_image_transform(
                options.get('resize_height'), options.get('resize_width'),
                options.get('resize_method') or 'bilinear')

        self._other_transforms = self._hparams.dataset.other_transformations

        if data_source is None:
            data_source = PickleDataSource[Dict[str, Any]](
                self._hparams.dataset.files)

        super().__init__(data_source, hparams, device)

    class _RecordWriter(io.BytesIO):
        def __init__(self, file_path: str,
                     features: Dict[str, FeatureDescription]):
            super().__init__()
            self._file_path = file_path
            self._features = features
            self._file_handle = open(self._file_path, 'wb')

        def close(self) -> None:
            self._file_handle.close()

        def write(self, example: Dict[str, Any]):  # type: ignore
            converted = {}
            for key, descriptor in self._features.items():
                value = example[key]
                if descriptor.collate_method is CollateMethod.List:
                    converted[key] = value
                    continue
                # Convert to NumPy array.
                value = np.asarray(value, dtype=descriptor.dtype)
                _check_shape(value, key, descriptor)
                converted[key] = value
            pickle.dump(converted, self._file_handle)

    @classmethod
    def writer(cls, file_path: str,
               feature_types: Dict[str, Tuple[Any, ...]]) \
            -> '_RecordWriter':
        r"""Construct a file writer object that saves records in pickled format.

        Example:

        .. code-block:: python

            file_path = "data/train.pkl"
            feature_types = {
                "input_ids": ["int64", "stacked_tensor", 128],
                "label_ids": ["int64", "stacked_tensor"],
            }
            with tx.data.RecordData.writer(file_path, feature_types) as writer:
                writer.write({
                    "input_ids": np.random.randint(0, 100, size=128),
                    "label_ids": np.random.randint(0, 100),
                })

        Args:
            file_path (str): Path to save the dataset.
            feature_types: Feature names and types. Please refer to
                :meth:`default_hparams` for details.

        Returns:
            A file writer object.
        """
        feature_types = _convert_feature_hparams(feature_types)
        return cls._RecordWriter(file_path, feature_types)

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of default hyperparameters.

        .. code-block:: python

            {
                # (1) Hyperparameters specific to the record data
                'dataset': {
                    'files': [],
                    'feature_types': {},
                    'feature_convert_types': {},
                    'image_options': {},
                    "num_shards": None,
                    "shard_id": None,
                    "other_transformations": [],
                    "data_name": None,
                }
                # (2) General hyperparameters
                "num_epochs": 1,
                "batch_size": 64,
                "allow_smaller_final_batch": True,
                "shuffle": True,
                "shuffle_buffer_size": None,
                "shard_and_shuffle": False,
                "num_parallel_calls": 1,
                "prefetch_buffer_size": 0,
                "max_dataset_size": -1,
                "seed": None,
                "name": "tfrecord_data",
            }

        Here:

        1. For the hyperparameters in the :attr:`"dataset"` field:

           `"files"`: str or list
               A (list of) pickled file path(s).

           `"feature_types"`: dict
               The feature names (`str`) with their descriptions in the form of
               ``feature_name: [dtype, feature_collate_method, shape]``:

               - ``dtype`` is a Python type (``int``, ``str``), dtype instance
                 from PyTorch (``torch.float``), NumPy (``np.int64``),
                 or TensorFlow (``tf.string``), or their stringified names such
                 as ``"torch.float"`` and ``"np.int64"``. The feature will be
                 read from the files and parsed into this dtype.

               - ``feature_collate_method`` is of type ``str``, and describes
                 how features are collated in the batch. Available values are:

                 - ``"stacked_tensor"``: Features are assumed to be tensors of a
                   fixed shape (or scalars). When collating, features are
                   stacked, with the batch dimension being the first dimension.
                   This is the default value if ``feature_collate_method`` is
                   not specified. For example:

                   - 5 scalar features -> a tensor of shape [5].
                   - 4 tensor features, each of shape [6, 5] -> a tensor of
                     shape [4, 6, 5].

                 - ``"padded_tensor"``: Features are assumed to be tensors, with
                   all dimensions except the first having the same size. When
                   collating, features are padded with zero values along the
                   end of the first dimension so that every tensor has the same
                   size, and then stacked, with the batch dimension being the
                   first dimension. For example:

                   - 3 tensor features, with shapes [4, 7, 8], [5, 7, 8], and
                     [4, 7, 8] -> a tensor of shape [3, 5, 7, 8].

                 - ``"list"``: Features can be any objects. When collating, the
                   features are stored in a Python list.

               - ``shape`` is optional, and can be of type ``int``, `tuple``, or
                 ``torch.Size``. If specified, shapes of tensor features will be
                 checked, depending on the ``feature_collate_method``:

                 - ``"stacked_tensor"``: The shape of every feature tensor must
                   be ``shape``.
                 - ``"padded_tensor"``: The shape (excluding first dimension)
                   of every feature tensor must be ``shape``.
                 - ``"list"``: ``shape`` is ignored.

                 .. note::
                    Shape check is performed before any transformations are
                    applied.

               Example:

               .. code-block:: python

                   feature_types = {
                       "input_ids": ["int64", "stacked_tensor", 128],
                       "label_ids": ["int64", "stacked_tensor"],
                       "name_lists": ["string", "list"],
                   }

               .. note::
                   This field is named `"feature_original_types"` in Texar-TF.
                   This name is still supported, but is deprecated in favor of
                   `"feature_types"`.

                   Texar-TF also uses different names for feature types:

                   - ``"FixedLenFeature"`` corresponds to ``"stacked_tensor"``.
                   - ``"FixedLenSequenceFeature"`` corresponds to
                     ``"padded_tensor"``.
                   - ``"VarLenFeature"`` corresponds to ``"list"``.

                   These names are also accepted in Texar-PyTorch, but are
                   deprecated in favor of the new names.

           `"feature_convert_types"`: dict, optional
               Specifies dtype converting after reading the data files. This
               `dict` maps feature names to desired data dtypes. For example,
               you can first read a feature into dtype ``torch.int32`` by
               specifying in :attr:`"feature_types"` above, and convert
               the feature to dtype ``"torch.long"`` by specifying here.
               Features not specified here will not do dtype-convert.

               - ``dtype`` is a Python type (`int`, `str`), dtype instance from
                 PyTorch (``torch.float``), NumPy (``np.int64``),
                 or TensorFlow (``tf.string``), or their stringified names such
                 as ``"torch.float"`` and ``"np.int64"``.

               Note that this converting process happens after all the data
               are restored.

               Example:

               .. code-block:: python

                   feature_convert_types = {
                       "input_ids": "int32",
                       "label_ids": "int32",
                   }

           `"image_options"`: dict, optional
               Specifies the image feature name and performs image resizing,
               includes three fields:

               - `"image_feature_name"`: str
                   The name of the feature which contains the image data. If
                   set, the image data will be restored in a `numpy.ndarray`.
               - `"resize_height"`: int
                   The height of the image after resizing.
               - `"resize_width"`: int
                   The width of the image after resizing.

               If any of :attr:`"resize_height"` or :attr:`"resize_width"` is
               not set, image data will be restored with original shape.

           `"num_shards"`: int, optional
               The number of data shards in distributed mode. Usually set to
               the number of processes in distributed computing.
               Used in combination with :attr:`"shard_id"`.

               .. warning::
                   Sharding is not yet supported. This option (and
                   related ones below) will be ignored.

           `"shard_id"`: int, optional
               Sets the unique id to identify a shard. The module will
               processes only the corresponding shard of the whole data.
               Used in combination with :attr:`"num_shards"`.

               For example, in a case of distributed computing on 2 GPUs, the
               hyperparameters of the data module for the two processes can be
               configured as below, respectively.

               For GPU 0:

               .. code-block:: python

                   dataset: {
                       ...
                       "num_shards": 2,
                       "shard_id": 0
                   }

               For GPU 1:

               .. code-block:: python

                   dataset: {
                       ...
                       "num_shards": 2,
                       "shard_id": 1
                   }

               Also refer to `examples/bert` for a use case.

           `"other_transformations"`: list
               A list of transformation functions or function names/paths to
               further transform each single data instance.

           `"data_name"`: str
               Name of the dataset.

        2. For the **general** hyperparameters, see
           :meth:`texar.torch.data.DatasetBase.default_hparams` for details.
        """
        hparams = DatasetBase.default_hparams()
        hparams["name"] = "record_data"
        hparams.update({
            "dataset": _default_record_dataset_hparams()
        })
        return hparams

    def process(self, raw_example: Dict[str, Any]) -> Dict[str, Any]:
        for key, descriptor in self._features.items():
            _check_shape(raw_example[key], key, descriptor)
        example = raw_example
        for key, dtype in self._convert_types.items():
            example[key] = np.asarray(example[key], dtype=dtype)
        for key, transform in self._image_transforms.items():
            example[key] = transform(example[key])
        for transform in self._other_transforms:
            example = transform(example)
        return example

    def collate(self, examples: List[Dict[str, Any]]) -> Batch:
        batch = {}
        for key, descriptor in self._features.items():
            values = [ex[key] for ex in examples]
            if descriptor.collate_method is not CollateMethod.List:
                # NumPy functions work on PyTorch tensors too.
                if descriptor.collate_method is CollateMethod.StackedTensor:
                    values = np.stack(values, axis=0)
                else:  # padded_tensor
                    values, _ = padded_batch(values)
                if (not isinstance(values, torch.Tensor) and
                        descriptor.dtype not in [np.str_, np.bytes_]):
                    values = torch.from_numpy(values)
            else:
                # Just put everything in a Python list.
                pass
            batch[key] = values
        return Batch(len(examples), batch)

    def list_items(self) -> List[str]:
        r"""Returns the list of item names that the data can produce.

        Returns:
            A list of strings.
        """
        return list(self._features.keys())

    @property
    def feature_names(self):
        r"""A list of feature names.
        """
        return self.list_items()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Custom initializers used in various methods.
"""

import math

import torch


def variance_scaling_initializer(inputs: torch.Tensor,
                                 factor: float = 2.0, mode: str = 'FAN_IN',
                                 uniform: bool = False):
    r"""Returns an initializer that generates tensors without scaling variance.
    When initializing a deep network, it is in principle advantageous to keep
    the scale of the input variance constant, so it does not explode or diminish
    by reaching the final layer. This initializer use the following formula:
    ```python
        if mode='FAN_IN': # Count only number of input connections.
        n = fan_in
        elif mode='FAN_OUT': # Count only number of output connections.
        n = fan_out
        elif mode='FAN_AVG': # Average number of inputs and output connections.
        n = (fan_in + fan_out)/2.0
        truncated_normal(shape, 0.0, stddev=sqrt(factor / n))
    ```
    * To get [Delving Deep into Rectifiers](
        http://arxiv.org/pdf/1502.01852v1.pdf) (also know as the "MSRA
        initialization"), use (Default):<br/>
        `factor=2.0 mode='FAN_IN' uniform=False`
    * To get [Convolutional Architecture for Fast Feature Embedding](
        http://arxiv.org/abs/1408.5093), use:<br/>
        `factor=1.0 mode='FAN_IN' uniform=True`
    * To get [Understanding the difficulty of training deep feed-forward
        neural networks](
        http://jmlr.org/proceedings/papers/v9/glorot10a/glorot10a.pdf),
        use:<br/>
        `factor=1.0 mode='FAN_AVG' uniform=True.`
    * To get `xavier_initializer` use either:<br/>
        `factor=1.0 mode='FAN_AVG' uniform=True`, or<br/>
        `factor=1.0 mode='FAN_AVG' uniform=False`.
    Args:
        factor: Float.  A multiplicative factor.
        mode: String.  'FAN_IN', 'FAN_OUT', 'FAN_AVG'.
        uniform: Whether to use uniform or normal distributed
                 random initialization.
    Returns:
        An initializer that generates tensors with unit variance.
    Raises:
        ValueError: if `dtype` is not a floating point type.
        TypeError: if `mode` is not in ['FAN_IN', 'FAN_OUT', 'FAN_AVG'].
    """

    # Estimating fan_in and fan_out is not possible to do perfectly, but we
    # try. This is the right thing for matrix multiply and convolutions.
    shape = inputs.size()
    fan_in = float(shape[-2]) if len(shape) > 1 else float(shape[-1])
    fan_out = float(shape[-1])
    for dim in shape[:-2]:
        fan_in *= float(dim)
        fan_out *= float(dim)

    if mode == 'FAN_IN':
        # Count only number of input connections.
        n = fan_in
    elif mode == 'FAN_OUT':
        # Count only number of output connections.
        n = fan_out
    elif mode == 'FAN_AVG':
        # Average number of inputs and output connections.
        n = (fan_in + fan_out) / 2.0
    else:
        raise ValueError(f"Unknown mode {mode} [FAN_IN, FAN_OUT, FAN_AVG]")

    if uniform:
        # To get stddev = math.sqrt(factor / n) need to
        # adjust for uniform.
        limit = math.sqrt(3.0 * factor / n)
        inputs.data.uniform_(-limit, limit)
    else:
        # To get stddev = math.sqrt(factor / n) need to
        # adjust for truncated normal.
        trunc_stddev = math.sqrt(1.3 * factor / n)

        u1 = torch.rand(shape) * (1 - math.exp(-2)) + math.exp(-2)
        u2 = torch.rand(shape)
        rnd = torch.sqrt(-2 * torch.log(u1)) * torch.cos(2 * math.pi * u2)

        ret = rnd * trunc_stddev
        inputs.data = ret



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
XLNet encoder.
"""

from typing import Any, Dict, List, Optional, Tuple, Union

import torch
from torch import nn
from torch.nn import functional as F

from texar.torch.modules.encoders.encoder_base import EncoderBase
from texar.torch.modules.pretrained.xlnet import PretrainedXLNetMixin
from texar.torch.modules.pretrained.xlnet_utils import (
    PositionWiseFF, RelativeMultiheadAttention, RelativePositionalEncoding,
    params_except_in)
from texar.torch.utils.utils import dict_fetch, sum_tensors

__all__ = [
    "XLNetEncoder",
]


class XLNetEncoder(EncoderBase, PretrainedXLNetMixin):
    r"""Raw XLNet module for encoding sequences. Please see
    :class:`~texar.torch.modules.PretrainedXLNetMixin` for a brief description
    of XLNet.

    Args:
        pretrained_model_name (optional): a `str`, the name
            of pre-trained model (e.g., ``xlnet-based-cased``). Please refer to
            :class:`~texar.torch.modules.PretrainedXLNetMixin` for
            all supported models.
            If `None`, the model name in :attr:`hparams` is used.
        cache_dir (optional): the path to a folder in which the
            pre-trained models will be cached. If `None` (default),
            a default directory (``texar_data`` folder under user's home
            directory) will be used.
        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparameter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure
            and default values.
    """
    _IS_DECODE = False

    def __init__(self,
                 pretrained_model_name: Optional[str] = None,
                 cache_dir: Optional[str] = None,
                 hparams=None):
        super().__init__(hparams=hparams)
        self.load_pretrained_config(pretrained_model_name, cache_dir)

        num_layers = self._hparams.num_layers
        num_heads = self._hparams.num_heads
        head_dim = self._hparams.head_dim

        self.word_embed = nn.Embedding(self._hparams.vocab_size,
                                       self._hparams.hidden_dim)
        self.pos_embed = RelativePositionalEncoding(
            hparams={
                "dim": self._hparams.hidden_dim,
                "max_seq_len": self._hparams.max_seq_length,
            })
        self.dropout = nn.Dropout(self._hparams.dropout)

        self.r_r_bias = None
        self.r_w_bias = None
        self.r_s_bias = None

        if not self._hparams.untie_r:
            self.r_r_bias = nn.Parameter(torch.Tensor(num_heads, head_dim))
            self.r_w_bias = nn.Parameter(torch.Tensor(num_heads, head_dim))
            self.r_s_bias = (nn.Parameter(torch.Tensor(num_heads, head_dim))
                             if self._hparams.use_segments else None)

        self.attn_layers = nn.ModuleList()
        self.ff_layers = nn.ModuleList()
        rel_attn_hparams = dict_fetch(
            self._hparams, RelativeMultiheadAttention.default_hparams())
        ff_hparams = dict_fetch(
            self._hparams, PositionWiseFF.default_hparams())
        for _ in range(num_layers):
            self.attn_layers.append(RelativeMultiheadAttention(
                self.r_r_bias, self.r_w_bias, self.r_s_bias,
                hparams=rel_attn_hparams))
            self.ff_layers.append(PositionWiseFF(hparams=ff_hparams))

        self.mask_emb = nn.Parameter(
            torch.Tensor(1, 1, self._hparams.hidden_dim))

        if self._IS_DECODE:
            self.lm_bias = nn.Parameter(torch.zeros(self._hparams.vocab_size))

        self.init_pretrained_weights()

    @staticmethod
    def default_hparams() -> Dict[str, Any]:
        r"""Returns a dictionary of hyperparameters with default values.

        * The encoder arch is determined by the constructor argument
          :attr:`pretrained_model_name` if it's specified. In this case,
          `hparams` are ignored.
        * Otherwise, the encoder arch is determined by
          `hparams['pretrained_model_name']` if it's specified. All other
          configurations in `hparams` are ignored.
        * If the above two are `None`, the encoder arch is defined by the
          configurations in `hparams` and weights are randomly initialized.

        .. code-block:: python

            {
                "pretrained_model_name": "xlnet-base-cased",
                "untie_r": True,
                "num_layers": 12,
                "mem_len": 0,
                "reuse_len": 0,
                "num_heads": 12,
                "hidden_dim": 768,
                "head_dim": 64,
                "dropout": 0.1,
                "attention_dropout": 0.1,
                "use_segments": True,
                "ffn_inner_dim": 3072,
                "activation": 'gelu',
                "vocab_size": 32000,
                "max_seq_length": 512,
                "initializer": None,
                "name": "xlnet_encoder",
            }

        Here:

        The default parameters are values for cased XLNet-Base model.

        `"pretrained_model_name"`: str or None
            The name of the pre-trained XLNet model. If None, the model
            will be randomly initialized.

        `"untie_r"`: bool
            Whether to untie the biases in attention.

        `"num_layers"`: int
            The number of stacked layers.

        `"mem_len"`: int
            The number of tokens to cache.

        `"reuse_len"`: int
            The number of tokens in the current batch to be cached and reused
            in the future.

        `"num_heads"`: int
            The number of attention heads.

        `"hidden_dim"`: int
            The hidden size.

        `"head_dim"`: int
            The dimension size of each attention head.

        `"dropout"`: float
            Dropout rate.

        `"attention_dropout"`: float
            Dropout rate on attention probabilities.

        `"use_segments"`: bool
            Whether to use segment embedding.

        `"ffn_inner_dim"`: int
            The hidden size in feed-forward layers.

        `"activation"`: str
            `relu` or `gelu`.

        `"vocab_size"`: int
            The vocabulary size.

        `"max_seq_length"`: int
            The maximum sequence length for `RelativePositionalEncoding`.

        `"initializer"`: dict, optional
            Hyperparameters of the default initializer that initializes
            variables created in this module.
            See :func:`~texar.torch.core.get_initializer` for details.

        `"name"`: str
            Name of the module.
        """

        return {
            'pretrained_model_name': 'xlnet-base-cased',
            'untie_r': True,
            'num_layers': 12,
            'mem_len': 0,
            'reuse_len': 0,
            # layer
            'num_heads': 12,
            'hidden_dim': 768,
            'head_dim': 64,
            'dropout': 0.1,
            'attention_dropout': 0.1,
            'use_segments': True,
            # ffn
            'ffn_inner_dim': 3072,
            'activation': 'gelu',
            # embedding
            'vocab_size': 32000,
            'max_seq_length': 512,
            'initializer': None,
            'name': "xlnet_encoder",
            '@no_typecheck': ['pretrained_model_name'],
        }

    def param_groups(self,
                     lr: Optional[float] = None,
                     lr_layer_scale: float = 1.0,
                     decay_base_params: bool = False):
        r"""Create parameter groups for optimizers. When
        :attr:`lr_layer_decay_rate` is not 1.0, parameters from each layer form
        separate groups with different base learning rates.

        The return value of this method can be used in the constructor of
        optimizers, for example:

        .. code-block:: python

            model = XLNetEncoder(...)
            param_groups = model.param_groups(lr=2e-5, lr_layer_scale=0.8)
            optim = torch.optim.Adam(param_groups)

        Args:
            lr (float): The learning rate. Can be omitted if
                :attr:`lr_layer_decay_rate` is 1.0.
            lr_layer_scale (float): Per-layer LR scaling rate. The `i`-th layer
                will be scaled by `lr_layer_scale ^ (num_layers - i - 1)`.
            decay_base_params (bool): If `True`, treat non-layer parameters
                (e.g. embeddings) as if they're in layer 0. If `False`, these
                parameters are not scaled.

        Returns:
            The parameter groups, used as the first argument for optimizers.
        """

        if lr_layer_scale != 1.0:
            if lr is None:
                raise ValueError(
                    "lr must be specified when lr_layer_decay_rate is not 1.0")

            num_layers = self._hparams.num_layers
            base_group = {
                "params": params_except_in(
                    self, ['attn_layers', 'ff_layers']),
                "lr": lr * (lr_layer_scale ** num_layers
                            if decay_base_params else 1.0)
            }
            param_groups = [base_group]
            for idx in range(num_layers):
                decay_rate = lr_layer_scale ** (num_layers - idx - 1)
                param_group = {
                    "params": [*self.attn_layers[idx].parameters(),
                               *self.ff_layers[idx].parameters()],
                    "lr": lr * decay_rate,
                }
                param_groups.append(param_group)
            return param_groups
        return self.parameters()

    @property
    def output_size(self):
        r"""The feature size of :meth:`forward` output.
        """
        return self._hparams.hidden_dim

    @staticmethod
    def _cache_mem(output: torch.Tensor,
                   prev_mem: Optional[torch.Tensor],
                   mem_len: int,
                   reuse_len: int = 0) -> torch.Tensor:
        r"""Cache hidden states into memory."""
        assert mem_len > 0

        if reuse_len is not None and reuse_len > 0:
            output = output[:reuse_len]
        if prev_mem is None:
            new_mem = output[-mem_len:]
        else:
            new_mem = torch.cat([prev_mem, output], dim=0)[-mem_len:]
        return new_mem.detach()

    def _create_causal_attn_mask(self,
                                 seq_len: int,
                                 mem_len: int,
                                 same_length: bool = False) -> torch.Tensor:
        r"""Create causal attention mask of shape
        `(seq_len, mem_len + seq_len)`.
        """
        assert self.r_w_bias is not None
        device = self.r_w_bias.device
        attn_mask = torch.ones(seq_len, seq_len, device=device)
        mask_u = torch.triu(attn_mask, diagonal=1)
        attn_mask_pad = torch.zeros(seq_len, mem_len, device=device)
        ret = torch.cat([attn_mask_pad, mask_u], dim=1)
        if same_length:
            mask_l = torch.tril(attn_mask, diagonal=-1)
            ret = torch.cat([ret[:, :seq_len] + mask_l, ret[:, seq_len:]], 1)
        return ret

    def forward(self,  # type: ignore
                inputs: Union[torch.Tensor, torch.LongTensor],
                segment_ids: Optional[torch.LongTensor] = None,
                input_mask: Optional[torch.Tensor] = None,
                memory: Optional[List[torch.Tensor]] = None,
                permute_mask: Optional[torch.Tensor] = None,
                target_mapping: Optional[torch.Tensor] = None,
                bi_data: bool = False,
                clamp_len: Optional[int] = None,
                cache_len: int = 0,
                same_length: bool = False,
                attn_type: str = 'bi',
                two_stream: bool = False) \
            -> Tuple[torch.Tensor, Optional[List[torch.Tensor]]]:
        r"""Compute XLNet representations for the input.

        Args:
            inputs: Either a **2D Tensor** of shape `[batch_size, max_time]`,
                containing the ids of tokens in input sequences, or
                a **3D Tensor** of shape `[batch_size, max_time, vocab_size]`,
                containing soft token ids (i.e., weights or probabilities)
                used to mix the embedding vectors.
            segment_ids: Shape `[batch_size, max_time]`.
            input_mask: Float tensor of shape `[batch_size, max_time]`. Note
                that positions with value 1 are masked out.
            memory: Memory from previous batches. A list of length `num_layers`,
                each tensor of shape `[batch_size, mem_len, hidden_dim]`.
            permute_mask: The permutation mask. Float tensor of shape
                `[batch_size, max_time, max_time]`.
                A value of 0 for ``permute_mask[i, j, k]`` indicates that
                position `i` attends to position `j` in batch `k`.
            target_mapping: The target token mapping. Float tensor of shape
                `[batch_size, num_targets, max_time]`.
                A value of 1 for ``target_mapping[i, j, k]`` indicates that
                the `i`-th target token (in order of permutation) in batch `k`
                is the token at position `j`.
                Each row ``target_mapping[i, :, k]`` can have no more than one
                value of 1.
            bi_data (bool): Whether to use bidirectional data input pipeline.
            clamp_len (int): Clamp all relative distances larger than
                :attr:`clamp_len`. A value of -1 means no clamping.
            cache_len (int): Length of memory (number of tokens) to cache.
            same_length (bool): Whether to use the same attention length for
                each token.
            attn_type (str): Attention type. Supported values are `"uni"`
                and `"bi"`.
            two_stream (bool): Whether to use two-stream attention. Only set to
                `True` when pre-training or generating text. Defaults to
                `False`.

        :returns: A tuple of `(output, new_memory)`:

            - **`output`**: The final layer output representations. Shape
              `[batch_size, max_time, hidden_dim]`.
            - **`new_memory`**: The memory of the current batch.
              If `cache_len` is 0, then `new_memory` is `None`. Otherwise, it is
              a list of length `num_layers`, each tensor of shape
              `[batch_size, cache_len, hidden_dim]`.
              This can be used as the :attr:`memory` argument in the next batch.
        """
        if inputs.dim() == 2:
            word_embeds = self.word_embed(inputs)
        elif inputs.dim() == 3:
            word_embeds = torch.tensordot(inputs, self.word_embed.weight,
                                          dims=([-1], [0]))
        else:
            raise ValueError("'inputs' should be a 2D or 3D tensor.")

        return self._forward(word_embed=word_embeds,
                             segment_ids=segment_ids,
                             input_mask=input_mask,
                             memory=memory,
                             permute_mask=permute_mask,
                             target_mapping=target_mapping,
                             bi_data=bi_data,
                             clamp_len=clamp_len,
                             cache_len=cache_len,
                             same_length=same_length,
                             attn_type=attn_type,
                             two_stream=two_stream)

    def _forward(self,
                 word_embed: torch.Tensor,
                 segment_ids: Optional[torch.LongTensor] = None,
                 input_mask: Optional[torch.Tensor] = None,
                 memory: Optional[List[torch.Tensor]] = None,
                 permute_mask: Optional[torch.Tensor] = None,
                 target_mapping: Optional[torch.Tensor] = None,
                 bi_data: bool = False,
                 clamp_len: Optional[int] = None,
                 cache_len: int = 0,
                 same_length: bool = False,
                 attn_type: str = 'bi',
                 two_stream: bool = False) \
            -> Tuple[torch.Tensor, Optional[List[torch.Tensor]]]:
        r"""Compute XLNet representations for the input. This layer exists
        because :class:`XLNetDecoder` compute embeddings in the decoder helper.
        `word_embed` has shape `[batch_size, max_time, word_embed_dim]`.
        Please refer to :meth:`forward` for the detailed information of other
        arguments.
        """
        # seq_len == max_time
        # word_embed: [seq_len, batch_size, word_embed_dim]
        word_embed = word_embed.permute(1, 0, 2)
        # segment_ids: [seq_len, batch_size]
        if segment_ids is not None:
            segment_ids = segment_ids.permute(1, 0)
        # input_mask: [seq_len, batch_size]
        if input_mask is not None:
            input_mask = input_mask.permute(1, 0)
        # memory: A list of length num_layers
        # each tensor of shape [mem_len, batch_size, hidden_dim]
        if memory is not None:
            memory = [m.permute(1, 0, 2) for m in memory]
        # permute_mask: [seq_len, seq_len, batch_size]
        if permute_mask is not None:
            permute_mask = permute_mask.permute(1, 2, 0)
        # target_mapping: [num_targets, seq_len, batch_size]
        if target_mapping is not None:
            target_mapping = target_mapping.permute(1, 2, 0)

        seq_len, batch_size = word_embed.size()[:2]
        mem_len = memory[0].size(0) if memory is not None else 0
        tot_len = seq_len + mem_len
        reuse_len = self._hparams.reuse_len

        # Construct masks.
        masks: List[Optional[torch.Tensor]] = []

        # Causal attention mask.
        if attn_type == 'uni':
            causal_mask = self._create_causal_attn_mask(
                seq_len, mem_len, same_length)
            # attn_mask: (seq_len, tot_len, 1, 1)
            causal_mask = causal_mask.unsqueeze(2).unsqueeze(3)
            masks.append(causal_mask)
        elif attn_type == 'bi':
            pass
        else:
            raise ValueError(f"Unsupported attention type: {attn_type}")

        # Data mask: input mask & permutation mask.
        if input_mask is not None:
            input_mask = input_mask.expand(seq_len, -1, -1)
        data_mask = sum_tensors([input_mask, permute_mask])
        if data_mask is not None:
            # All positions in memory can be attended to.
            memory_mask = data_mask.new_zeros(seq_len, mem_len, batch_size)
            # data_mask: (seq_len, tot_len, batch_size, 1)
            data_mask = torch.cat([memory_mask, data_mask], dim=1).unsqueeze(3)
            masks.append(data_mask)

        # Exclude the main diagonal (target tokens) from the mask.
        attn_mask = sum_tensors(masks)
        if attn_mask is None:
            final_mask = None
        else:
            attn_mask = (attn_mask > 0)
            final_mask = -torch.eye(seq_len, device=attn_mask.device)
            final_mask = torch.cat([
                final_mask.new_zeros(seq_len, mem_len), final_mask], dim=-1)
            final_mask = final_mask.unsqueeze(2).unsqueeze(3)
            # final_mask: (seq_len, tot_len, batch_size, 1)
            final_mask = ((attn_mask.float() + final_mask) > 0)

        # Construct segment embedding.
        if segment_ids is not None:
            concat_segment_ids = torch.cat([
                segment_ids.new_zeros(mem_len, batch_size), segment_ids])
            segment_matrix = (segment_ids.unsqueeze(1) !=
                              concat_segment_ids.unsqueeze(0)).long()
            segment_matrix = F.one_hot(segment_matrix, num_classes=2).float()
        else:
            segment_matrix = None

        pos_embed = self.pos_embed(
            batch_size, seq_len, tot_len, clamp_len, attn_type, bi_data)
        pos_embed = self.dropout(pos_embed)

        states_h = self.dropout(word_embed)
        states_g = None
        if two_stream:
            if target_mapping is not None:
                word_embed_q = self.mask_emb.expand(
                    target_mapping.size(0), batch_size, -1)
            else:
                word_embed_q = word_embed
            states_g = self.dropout(word_embed_q)
        new_memory = []

        for idx in range(self._hparams.num_layers):
            cur_memory = memory[idx] if memory is not None else None
            if cache_len > 0:
                new_memory.append(self._cache_mem(
                    states_h, cur_memory, cache_len, reuse_len))
            attn_layer: RelativeMultiheadAttention
            attn_layer = self.attn_layers[idx]  # type: ignore
            states_h, states_g = attn_layer(
                states_h=states_h, states_g=states_g,
                pos_embed=pos_embed, segment_mat=segment_matrix,
                attn_mask_h=final_mask, attn_mask_g=attn_mask,
                target_mapping=target_mapping, memory=cur_memory)
            states_h = self.ff_layers[idx](states_h)
            if states_g is not None:
                states_g = self.ff_layers[idx](states_g)

        output = self.dropout(states_h if states_g is None else states_g)

        # Now output: [seq_len, batch_size, hidden_dim]
        # new_memory: None or A list of length num_layers,
        # each tensor of shape [cache_len, batch_size, hidden_dim]
        output = output.permute(1, 0, 2)
        if new_memory is not None:
            new_memory = [m.permute(1, 0, 2) for m in new_memory]

        if cache_len == 0:
            return output, None

        return output, new_memory



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for data related operations.
"""
import tempfile
import copy
import numpy as np

import unittest

from texar.torch.data.data.data_iterators import DataIterator
from texar.torch.data.data.paired_text_data import PairedTextData
from texar.torch.data.vocabulary import SpecialTokens


class PairedTextDataTest(unittest.TestCase):
    """Tests paired text data class.
    """

    def setUp(self):

        # Create test data
        vocab_list = ['This', 'is', 'a', 'word', '词']
        vocab_file = tempfile.NamedTemporaryFile()
        vocab_file.write('\n'.join(vocab_list).encode("utf-8"))
        vocab_file.flush()
        self._vocab_file = vocab_file
        self._vocab_size = len(vocab_list)

        src_text = ['This is a sentence from source .', '词 词 。 source']
        src_text_file = tempfile.NamedTemporaryFile()
        src_text_file.write('\n'.join(src_text).encode("utf-8"))
        src_text_file.flush()
        self._src_text_file = src_text_file

        tgt_text = ['This is a sentence from target .', '词 词 。 target']
        tgt_text_file = tempfile.NamedTemporaryFile()
        tgt_text_file.write('\n'.join(tgt_text).encode("utf-8"))
        tgt_text_file.flush()
        self._tgt_text_file = tgt_text_file

        self._hparams = {
            "num_epochs": 1,
            "batch_size": 3,
            "source_dataset": {
                "files": [self._src_text_file.name],
                "vocab_file": self._vocab_file.name,
            },
            "target_dataset": {
                "files": self._tgt_text_file.name,
                "vocab_share": True,
                "eos_token": "<TARGET_EOS>"
            }
        }

        self.src_upper_cased_text = []
        for sent in src_text:
            upper_cased_tokens = sent.upper().split(" ")
            upper_cased_tokens.append(SpecialTokens.EOS)
            self.src_upper_cased_text.append(upper_cased_tokens)

        max_length = max([len(tokens) for tokens in self.src_upper_cased_text])
        self.src_upper_cased_text = [sent + [''] * (max_length - len(sent)) for
                                     sent in self.src_upper_cased_text]

        self.tgt_upper_cased_text = []
        for sent in tgt_text:
            upper_cased_tokens = sent.upper().split(" ")
            upper_cased_tokens.insert(0, "<BOS>")
            upper_cased_tokens.append("<TARGET_EOS>")
            self.tgt_upper_cased_text.append(upper_cased_tokens)

        max_length = max([len(tokens) for tokens in self.tgt_upper_cased_text])
        self.tgt_upper_cased_text = [sent + [''] * (max_length - len(sent)) for
                                     sent in self.tgt_upper_cased_text]

    def _run_and_test(self, hparams, proc_shr=False, test_transform=None,
                      discard_src=False):
        # Construct database
        text_data = PairedTextData(hparams)
        self.assertEqual(text_data.source_vocab.size,
                         self._vocab_size +
                         len(text_data.source_vocab.special_tokens))

        iterator = DataIterator(text_data)
        for data_batch in iterator:
            self.assertEqual(set(data_batch.keys()),
                             set(text_data.list_items()))

            if proc_shr:
                tgt_eos = '<EOS>'
            else:
                tgt_eos = '<TARGET_EOS>'

            # Test matching
            src_text = data_batch['source_text']
            tgt_text = data_batch['target_text']
            if proc_shr:
                for src, tgt in zip(src_text, tgt_text):
                    np.testing.assert_array_equal(src[:3], tgt[:3])
            else:
                for src, tgt in zip(src_text, tgt_text):
                    np.testing.assert_array_equal(src[:3], tgt[1:4])
            self.assertTrue(
                tgt_eos in data_batch['target_text'][0])

            if test_transform:
                for i in range(len(data_batch['source_text'])):
                    text_ = data_batch['source_text'][i]
                    self.assertTrue(text_ in self.src_upper_cased_text)
                for i in range(len(data_batch['target_text'])):
                    text_ = data_batch['target_text'][i]
                    self.assertTrue(text_ in self.tgt_upper_cased_text)

            if discard_src:
                src_hparams = text_data.hparams.source_dataset
                max_l = src_hparams.max_seq_length
                max_l += sum(int(x is not None and x != '')
                             for x in [text_data._src_bos_token,
                                       text_data._tgt_bos_token])
                for l in data_batch["source_length"]:
                    self.assertLessEqual(l, max_l)

    def test_default_setting(self):
        """Tests the logics of the text data.
        """
        self._run_and_test(self._hparams)

    def test_shuffle(self):
        """Tests toggling shuffle.
        """
        hparams = copy.copy(self._hparams)
        hparams["shuffle"] = False
        self._run_and_test(hparams)

    def test_processing_share(self):
        """Tests sharing processing.
        """
        hparams = copy.copy(self._hparams)
        hparams["target_dataset"]["processing_share"] = True
        self._run_and_test(hparams, proc_shr=True)

    def test_other_transformations(self):
        """Tests use of other transformations
        """
        _upper_func = lambda raw_example: [x.upper() for x in raw_example]

        hparams = copy.copy(self._hparams)
        hparams["source_dataset"].update(
            {"other_transformations": [_upper_func]})
        hparams["target_dataset"].update(
            {"other_transformations": [_upper_func]})
        self._run_and_test(hparams, test_transform=True)

    def test_length_filter(self):
        """Tests filtering by length.
        """
        hparams = copy.copy(self._hparams)
        hparams["source_dataset"].update(
            {"max_seq_length": 4,
             "length_filter_mode": "discard"})
        self._run_and_test(hparams, discard_src=True)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various embedders.
"""
from typing import Optional

import torch
from torch.nn import functional as F

from texar.torch.modules.embedders import embedder_utils
from texar.torch.modules.embedders.embedder_base import (
    EmbedderBase, EmbeddingDropout)

__all__ = [
    "WordEmbedder",
]


class WordEmbedder(EmbedderBase):
    r"""Simple word embedder that maps indexes into embeddings. The indexes
    can be soft (e.g., distributions over vocabulary).

    Either :attr:`init_value` or :attr:`vocab_size` is required. If both are
    given, there must be ``init_value.shape[0]==vocab_size``.

    Args:
        init_value (optional): A Tensor or numpy array that contains the
            initial value of embeddings. It is typically of shape
            ``[vocab_size] + embedding-dim``. Embeddings can have dimensionality
            > 1.

            If `None`, embedding is initialized as specified in
            ``hparams["initializer"]``. Otherwise, the
            ``"initializer"`` and ``"dim"`` hyperparameters in :attr:`hparams`
            are ignored.
        vocab_size (int, optional): The vocabulary size. Required if
            :attr:`init_value` is not given.
        hparams (dict, optional): Embedder hyperparameters. Missing
            hyperparameters will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure and
            default values.

    See :meth:`forward` for the inputs and outputs of the embedder.

    Example:

    .. code-block:: python

        ids = torch.empty([32, 10]).uniform_(to=10).type(torch.int64).
        soft_ids = torch.empty([32, 10, 100]).uniform_()

        embedder = WordEmbedder(vocab_size=100, hparams={'dim': 256})
        ids_emb = embedder(ids=ids) # shape: [32, 10, 256]
        soft_ids_emb = embedder(soft_ids=soft_ids) # shape: [32, 10, 256]

    .. code-block:: python

        # Use with Texar data module
        hparams={
            'dataset': {
                'embedding_init': {'file': 'word2vec.txt'}
                ...
            },
        }
        data = MonoTextData(data_params)
        iterator = DataIterator(data)
        batch = next(iter(iterator))

        # Use data vocab size
        embedder_1 = WordEmbedder(vocab_size=data.vocab.size)
        emb_1 = embedder_1(batch['text_ids'])

        # Use pre-trained embedding
        embedder_2 = WordEmbedder(init_value=data.embedding_init_value)
        emb_2 = embedder_2(batch['text_ids'])


    .. document private functions
    """

    def __init__(self, init_value: Optional[torch.Tensor] = None,
                 vocab_size: Optional[int] = None, hparams=None):

        if init_value is None and vocab_size is None:
            raise ValueError(
                "Either `init_value` or `vocab_size` is required.")

        super().__init__(init_value=init_value,
                         num_embeds=vocab_size, hparams=hparams)

        if vocab_size is None:
            self._vocab_size = self._num_embeds
        else:
            self._vocab_size = vocab_size
        if self._vocab_size != self._num_embeds:
            raise ValueError(
                f"vocab_size must equal to init_value.shape[0]. "
                f"Got {self._vocab_size} and {self._num_embeds}")

        self._dropout_layer = EmbeddingDropout(self._hparams.dropout_rate)

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "dim": 100,
                "dropout_rate": 0,
                "dropout_strategy": 'element',
                "initializer": {
                    "type": "random_uniform_initializer",
                    "kwargs": {
                        "minval": -0.1,
                        "maxval": 0.1,
                        "seed": None
                    }
                },
                "trainable": True,
                "name": "word_embedder",
            }

        Here:

        `"dim"`: int or list
            Embedding dimension. Can be a list of integers to yield embeddings
            with dimensionality > 1.

            Ignored if :attr:`init_value` is given to the embedder constructor.

        `"dropout_rate"`: float
            The dropout rate between 0 and 1. For example, ``dropout_rate=0.1``
            would zero out 10% of the embeddings. Set to 0 to disable dropout.

        `"dropout_strategy"`: str
            The dropout strategy. Can be one of the following

            - ``"element"``: The regular strategy that drops individual elements
              in the embedding vectors.
            - ``"item"``: Drops individual items (e.g., words) entirely. For
              example, for the word sequence "the simpler the better", the
              strategy can yield "_ simpler the better", where the first "the"
              is dropped.
            - ``"item_type"``: Drops item types (e.g., word types). For example,
              for the above sequence, the strategy can yield "_ simpler _
              better", where the word type "the" is dropped. The dropout will
              never yield "_ simpler the better" as in the ``"item"`` strategy.

        `"initializer"`: dict or None
            Hyperparameters of the initializer for embedding values. See
            :func:`~texar.torch.core.get_initializer` for the details. Ignored
            if :attr:`init_value` is given to the embedder constructor.

        `"trainable"`: bool
            Whether the embedding parameters are trainable. If false, freeze the
            embedding parameters.

        `"name"`: str
            Name of the embedding variable.
        """
        hparams = embedder_utils.default_embedding_hparams()
        hparams["name"] = "word_embedder"
        return hparams

    def extra_repr(self) -> str:
        return f"vocab_size={self.vocab_size}, embedding_dim={self.dim}"

    def forward(self,  # type: ignore
                ids: Optional[torch.LongTensor] = None,
                soft_ids: Optional[torch.Tensor] = None,
                **kwargs) -> torch.Tensor:
        r"""Embeds (soft) ids.

        Either :attr:`ids` or :attr:`soft_ids` must be given, and they
        must not be given at the same time.

        Args:
            ids (optional): An integer tensor containing the ids to embed.
            soft_ids (optional): A tensor of weights (probabilities) used to
                mix the embedding vectors.
            kwargs: Additional keyword arguments for
                :torch_nn:`functional.embedding` besides :attr:`params` and
                :attr:`ids`.

        Returns:
            If :attr:`ids` is given, returns a Tensor of shape
            ``list(ids.shape) + embedding-dim``. For example,
            if ``list(ids.shape) == [batch_size, max_time]``
            and ``list(embedding.shape) == [vocab_size, emb_dim]``, then the
            return tensor has shape ``[batch_size, max_time, emb_dim]``.

            If :attr:`soft_ids` is given, returns a Tensor of shape
            ``list(soft_ids.shape)[:-1] + embedding-dim``. For example,
            if ``list(soft_ids.shape) == [batch_size, max_time, vocab_size]``
            and ``list(embedding.shape) == [vocab_size, emb_dim]``, then the
            return tensor has shape ``[batch_size, max_time, emb_dim]``.
        """
        if ids is not None:
            if soft_ids is not None:
                raise ValueError(
                    'Must not specify `ids` and `soft_ids` at the same time.')
            ids_rank = ids.dim()
        elif soft_ids is not None:
            ids_rank = soft_ids.dim() - 1
        else:
            raise ValueError('Either `ids` or `soft_ids` must be given.')

        embedding = self._embedding

        if self._hparams.dropout_strategy == 'item_type':
            noise_shape = self._get_noise_shape(self._hparams.dropout_strategy)
            embedding = self._dropout_layer(embedding, noise_shape)

        if ids is not None:
            outputs = F.embedding(ids, embedding, **kwargs)
        else:
            outputs = embedder_utils.soft_embedding_lookup(embedding, soft_ids)

        if self._hparams.dropout_strategy != 'item_type':
            noise_shape = self._get_noise_shape(
                self._hparams.dropout_strategy,
                ids_rank=ids_rank, dropout_input=outputs)
            outputs = self._dropout_layer(outputs, noise_shape)

        return outputs

    @property
    def embedding(self) -> torch.Tensor:
        r"""The embedding tensor, of shape ``[vocab_size] + dim``.
        """
        return self._embedding

    @property
    def dim(self) -> int:
        r"""The embedding dimension.
        """
        return self._dim

    @property
    def vocab_size(self) -> int:
        r"""The vocabulary size.
        """
        return self._vocab_size

    @property
    def num_embeddings(self) -> int:
        r"""The vocabulary size. This interface matches
        :torch_nn:`Embedding`.
        """
        return self._vocab_size

    @property
    def output_size(self) -> int:
        r"""The feature size of :meth:`forward` output. If the :attr:`dim`
        hyperparameter is a ``list`` or ``tuple``, the feature size
        equals its final dimension; otherwise, if :attr:`dim` is an
        ``int``, the feature size equals :attr:`dim`.
        """
        if isinstance(self._dim, (list, tuple)):
            return self._dim[-1]
        else:
            return self._dim



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for bleu_tool.
"""

import unittest

import tempfile

from texar.torch.evals.bleu_transformer import (
    bleu_transformer_tokenize, file_bleu)


class BLEUToolTest(unittest.TestCase):
    r"""Test bleu_tool.
    """

    def _create_temp_file(self, text):
        temp_file = tempfile.NamedTemporaryFile(delete=False)
        with open(temp_file.name, "w") as f:
            f.write(text)
        return temp_file.name

    def test_bleu_same(self):
        ref = self._create_temp_file("test 1 two 3\nmore tests!")
        hyp = self._create_temp_file("test 1 two 3\nmore tests!")

        uncased_score = file_bleu(ref, hyp, case_sensitive=False)
        cased_score = file_bleu(ref, hyp, case_sensitive=True)
        self.assertEqual(100, uncased_score)
        self.assertEqual(100, cased_score)

    def test_bleu_same_different_case(self):
        ref = self._create_temp_file("Test 1 two 3\nmore tests!")
        hyp = self._create_temp_file("test 1 two 3\nMore tests!")
        uncased_score = file_bleu(ref, hyp, case_sensitive=False)
        cased_score = file_bleu(ref, hyp, case_sensitive=True)
        self.assertEqual(100, uncased_score)
        self.assertLess(cased_score, 100)

    def test_bleu_different(self):
        ref = self._create_temp_file("Testing\nmore tests!")
        hyp = self._create_temp_file("Dog\nCat")
        uncased_score = file_bleu(ref, hyp, case_sensitive=False)
        cased_score = file_bleu(ref, hyp, case_sensitive=True)
        self.assertLess(uncased_score, 100)
        self.assertLess(cased_score, 100)

    def test_bleu_tokenize(self):
        s = "Test0, 1 two, 3"
        tokenized = bleu_transformer_tokenize(s)
        self.assertEqual(["Test0", ",", "1", "two", ",", "3"], tokenized)

    def test_bleu_version(self):
        ref = self._create_temp_file("Test 1 two 3\nmore tests!")
        hyp = self._create_temp_file("test 1 two 3\nMore tests!")
        uncased_score = file_bleu(
            ref, hyp, bleu_version="corpus_bleu", case_sensitive=False)
        cased_score = file_bleu(
            ref, hyp, bleu_version="corpus_bleu", case_sensitive=True)
        self.assertEqual(100, uncased_score)
        self.assertLess(cased_score, 100)

        uncased_score = file_bleu(
            ref, hyp, bleu_version="corpus_bleu_moses", case_sensitive=False)
        cased_score = file_bleu(
            ref, hyp, bleu_version="corpus_bleu_moses", case_sensitive=True)
        self.assertEqual(100, uncased_score)
        self.assertLess(cased_score, 100)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Data consisting of multiple aligned parts.
"""
from typing import (Any, Callable, Dict, List, Optional, Tuple, Union)

import torch
from texar.torch.data.data.data_base import (
    DatasetBase, DataSource,
    FilterDataSource, ZipDataSource, SequenceDataSource)
from texar.torch.data.data.dataset_utils import Batch, connect_name
from texar.torch.data.data.mono_text_data import (
    MonoTextData, _LengthFilterMode, _default_mono_text_dataset_hparams)
from texar.torch.data.data.record_data import (
    PickleDataSource, RecordData, _default_record_dataset_hparams)
from texar.torch.data.data.scalar_data import (
    ScalarData, _default_scalar_dataset_hparams)
from texar.torch.data.data.text_data_base import (
    TextDataBase, TextLineDataSource)
from texar.torch.data.embedding import Embedding
from texar.torch.data.vocabulary import SpecialTokens, Vocab
from texar.torch.hyperparams import HParams
from texar.torch.utils import utils, dict_fetch
from texar.torch.utils.dtypes import is_str, get_supported_scalar_types

__all__ = [
    "_default_dataset_hparams",
    "MultiAlignedData"
]


def _is_text_data(data_type):
    return data_type == "text"


def _is_scalar_data(data_type):
    return data_type in get_supported_scalar_types()


def _is_record_data(data_type):
    return data_type == "record"


def _default_dataset_hparams(data_type=None):
    r"""Returns hyperparameters of a dataset with default values.

    See :meth:`texar.torch.data.MultiAlignedData.default_hparams` for details.
    """
    if data_type is None:
        data_type = "text"

    if _is_text_data(data_type):
        hparams = _default_mono_text_dataset_hparams()
        hparams.update({
            "data_type": data_type,
            "vocab_share_with": None,
            "embedding_init_share_with": None,
            "processing_share_with": None,
        })
    elif _is_scalar_data(data_type):
        hparams = _default_scalar_dataset_hparams()
    elif _is_record_data(data_type):
        hparams = _default_record_dataset_hparams()
        hparams.update({
            "data_type": data_type,
        })
    else:
        raise ValueError(f"Invalid data type '{data_type}'")
    return hparams


class MultiAlignedData(
    TextDataBase[Tuple[Union[str, Dict[str, Any]], ...],
                 Tuple[Union[List[str], int, float, Dict[str, Any]], ...]]):
    r"""Data consisting of multiple aligned parts.

    Args:
        hparams (dict): Hyperparameters. See :meth:`default_hparams` for the
            defaults.
        device: The device of the produced batches. For GPU training, set to
            current CUDA device.

    The processor can read any number of parallel fields as specified in
    the "datasets" list of :attr:`hparams`, and result in a Dataset whose
    element is a python `dict` containing data fields from each of the
    specified datasets. Fields from a text dataset or Record dataset have
    names prefixed by its :attr:`"data_name"`. Fields from a scalar dataset are
    specified by its :attr:`"data_name"`.

    Example:

        .. code-block:: python

            hparams={
                'datasets': [
                    {'files': 'a.txt', 'vocab_file': 'v.a', 'data_name': 'x'},
                    {'files': 'b.txt', 'vocab_file': 'v.b', 'data_name': 'y'},
                    {'files': 'c.txt', 'data_type': 'int', 'data_name': 'z'}
                ]
                'batch_size': 1
            }
            data = MultiAlignedData(hparams)
            iterator = DataIterator(data)

            for batch in iterator:
                # batch contains the following
                # batch == {
                #    'x_text': [['<BOS>', 'x', 'sequence', '<EOS>']],
                #    'x_text_ids': [['1', '5', '10', '2']],
                #    'x_length': [4]
                #    'y_text': [['<BOS>', 'y', 'sequence', '1', '<EOS>']],
                #    'y_text_ids': [['1', '6', '10', '20', '2']],
                #    'y_length': [5],
                #    'z': [1000],
                # }

            ...

            hparams={
                'datasets': [
                    {'files': 'd.txt', 'vocab_file': 'v.d', 'data_name': 'm'},
                    {
                        'files': 'd.tfrecord',
                        'data_type': 'tf_record',
                        "feature_types": {
                            'image': ['tf.string', 'stacked_tensor']
                        },
                        'image_options': {
                            'image_feature_name': 'image',
                            'resize_height': 512,
                            'resize_width': 512,
                        },
                        'data_name': 't',
                    }
                ]
                'batch_size': 1
            }
            data = MultiAlignedData(hparams)
            iterator = DataIterator(data)
            for batch in iterator:
                # batch contains the following
                # batch_ == {
                #    'x_text': [['<BOS>', 'NewYork', 'City', 'Map', '<EOS>']],
                #    'x_text_ids': [['1', '100', '80', '65', '2']],
                #    'x_length': [5],
                #
                #    # "t_image" is a list of a "numpy.ndarray" image
                #    # in this example. Its width is equal to 512 and
                #    # its height is equal to 512.
                #    't_image': [...]
                # }

    """

    def __init__(self, hparams, device: Optional[torch.device] = None):
        self._hparams = HParams(hparams, self.default_hparams())
        # Defaultizes hyperparameters of each dataset
        datasets_hparams = self._hparams.datasets
        defaultized_datasets_hparams = []
        for hparams_i in datasets_hparams:
            data_type = hparams_i.get("data_type", None)
            defaultized_ds_hpms = HParams(hparams_i,
                                          _default_dataset_hparams(data_type))
            defaultized_datasets_hparams.append(defaultized_ds_hpms)
        self._hparams.datasets = defaultized_datasets_hparams

        self._vocab = self.make_vocab(self._hparams.datasets)
        self._embedding = self.make_embedding(
            self._hparams.datasets, self._vocab)

        dummy_source = SequenceDataSource[Any]([])
        name_prefix: List[str] = []
        self._names: List[Dict[str, Any]] = []
        sources: List[DataSource] = []
        filters: List[Optional[Callable[[str], bool]]] = []
        self._databases: List[DatasetBase] = []
        for idx, hparams_i in enumerate(self._hparams.datasets):
            data_type = hparams_i.data_type
            source_i: DataSource

            if _is_text_data(data_type):
                source_i = TextLineDataSource(
                    hparams_i.files,
                    compression_type=hparams_i.compression_type,
                    delimiter=hparams_i.delimiter)
                sources.append(source_i)
                if ((hparams_i.length_filter_mode ==
                     _LengthFilterMode.DISCARD.value) and
                        hparams_i.max_seq_length is not None):

                    def _get_filter(max_seq_length):
                        return lambda x: len(x) <= max_seq_length

                    filters.append(_get_filter(hparams_i.max_seq_length))
                else:
                    filters.append(None)

                self._names.append({
                    field: connect_name(hparams_i.data_name, field)
                    for field in ["text", "text_ids", "length"]
                })

                dataset_hparams = dict_fetch(
                    hparams_i, MonoTextData.default_hparams()["dataset"])
                dataset_hparams["data_name"] = None
                self._databases.append(MonoTextData(
                    hparams={"dataset": dataset_hparams}, device=device,
                    vocab=self._vocab[idx],
                    embedding=self._embedding[idx],
                    data_source=dummy_source))
            elif _is_scalar_data(data_type):
                source_i = TextLineDataSource(
                    hparams_i.files,
                    compression_type=hparams_i.compression_type)
                sources.append(source_i)
                filters.append(None)
                self._names.append({"data": hparams_i.data_name})

                dataset_hparams = dict_fetch(
                    hparams_i, ScalarData.default_hparams()["dataset"])
                dataset_hparams["data_name"] = "data"
                self._databases.append(ScalarData(
                    hparams={"dataset": dataset_hparams}, device=device,
                    data_source=dummy_source))
            elif _is_record_data(data_type):
                source_i = PickleDataSource(file_paths=hparams_i.files)
                sources.append(source_i)
                # TODO: Only check `feature_types` when we finally remove
                #   `feature_original_types`.
                feature_types = (hparams_i.feature_types or
                                 hparams_i.feature_original_types)
                self._names.append({
                    name: connect_name(hparams_i.data_name, name)
                    for name in feature_types.keys()})
                filters.append(None)

                dataset_hparams = dict_fetch(
                    hparams_i, RecordData.default_hparams()["dataset"])
                self._databases.append(RecordData(
                    hparams={"dataset": dataset_hparams}, device=device,
                    data_source=dummy_source))
            else:
                raise ValueError(f"Unknown data type: {hparams_i.data_type}")

            # check for duplicate names
            for i in range(1, len(name_prefix)):
                if name_prefix[i] in name_prefix[:i - 1]:
                    raise ValueError(f"Duplicate data name: {name_prefix[i]}")

            name_prefix.append(hparams_i["data_name"])

        self._name_to_id = {v: k for k, v in enumerate(name_prefix)}

        data_source: DataSource = ZipDataSource(*sources)

        if any(filters):
            def filter_fn(data):
                return all(fn(data) for fn, data in zip(filters, data)
                           if fn is not None)

            data_source = FilterDataSource(data_source, filter_fn=filter_fn)
        super().__init__(data_source, self._hparams, device)

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of default hyperparameters:

        .. code-block:: python

            {
                # (1) Hyperparams specific to text dataset
                "datasets": []
                # (2) General hyperparams
                "num_epochs": 1,
                "batch_size": 64,
                "allow_smaller_final_batch": True,
                "shuffle": True,
                "shuffle_buffer_size": None,
                "shard_and_shuffle": False,
                "num_parallel_calls": 1,
                "prefetch_buffer_size": 0,
                "max_dataset_size": -1,
                "seed": None,
                "name": "multi_aligned_data",
            }

        Here:

        1. "datasets" is a list of `dict` each of which specifies a
           dataset which can be text, scalar or Record. The :attr:`"data_name"`
           field of each dataset is used as the name prefix of the data fields
           from the respective dataset. The :attr:`"data_name"` field of each
           dataset should not be the same.

           i) For scalar dataset, the allowed hyperparameters and default
              values are the same as the "dataset" field of
              :meth:`texar.torch.data.ScalarData.default_hparams`. Note that
              :attr:`"data_type"` must be explicitly specified
              (either "int" or "float").

           ii) For Record dataset, the allowed hyperparameters and default
               values are the same as the "dataset" field of
               :meth:`texar.torch.data.RecordData.default_hparams`. Note that
               :attr:`"data_type"` must be explicitly specified ("record").

           iii) For text dataset, the allowed hyperparameters and default
                values are the same as the "dataset" filed of
                :meth:`texar.torch.data.MonoTextData.default_hparams`, with
                several extra hyperparameters:

                `"data_type"`: str
                    The type of the dataset, one of {"text", "int", "float",
                    "record"}. If set to "int" or "float", the dataset is
                    considered to be a scalar dataset. If set to
                    "record", the dataset is considered to be a Record
                    dataset.

                    If not specified or set to "text", the dataset is
                    considered to be a text dataset.

                `"vocab_share_with"`: int, optional
                    Share the vocabulary of a preceding text dataset with
                    the specified index in the list (starting from 0). The
                    specified dataset must be a text dataset, and must have
                    an index smaller than the current dataset.

                    If specified, the vocab file of current dataset is
                    ignored. Default is `None` which disables the vocab
                    sharing.

                `"embedding_init_share_with"`: int, optional
                    Share the embedding initial value of a preceding text
                    dataset with the specified index in the list (starting
                    from 0). The specified dataset must be a text dataset,
                    and must have an index smaller than the current dataset.

                    If specified, the :attr:`"embedding_init"` field of the
                    current dataset is ignored. Default is `None` which
                    disables the initial value sharing.

                `"processing_share_with"`: int, optional
                    Share the processing configurations of a preceding text
                    dataset with the specified index in the list (starting
                    from 0). The specified dataset must be a text dataset,
                    and must have an index smaller than the current dataset.

                    If specified, relevant field of the current dataset are
                    ignored, including `delimiter`, `bos_token`,
                    `eos_token`, and "other_transformations". Default is
                    `None` which disables the processing sharing.

        2. For the **general** hyperparameters, see
        :meth:`texar.torch.data.DatasetBase.default_hparams` for details.

        """
        hparams = TextDataBase.default_hparams()
        hparams["name"] = "multi_aligned_data"
        hparams["datasets"] = []
        return hparams

    def to(self, device: Optional[torch.device]):
        for dataset in self._databases:
            dataset.to(device)
        return super().to(device)

    @staticmethod
    def _raise_sharing_error(err_data, share_data, hparam_name):
        raise ValueError(
            f"Must only share specifications with a preceding dataset. "
            f"Dataset {err_data} has '{hparam_name}={share_data:d}'")

    @staticmethod
    def make_vocab(hparams: List[HParams]) -> List[Optional[Vocab]]:
        r"""Makes a list of vocabs based on the hyperparameters.

        Args:
            hparams (list): A list of dataset hyperparameters.

        Returns:
            A list of :class:`texar.torch.data.Vocab` instances. Some instances
            may be the same objects if they are set to be shared and have
            the same other configurations.
        """
        vocabs: List[Optional[Vocab]] = []
        for i, hparams_i in enumerate(hparams):
            if not _is_text_data(hparams_i.data_type):
                vocabs.append(None)
                continue

            proc_share = hparams_i.processing_share_with
            if proc_share is not None:
                bos_token = hparams[proc_share].bos_token
                eos_token = hparams[proc_share].eos_token
            else:
                bos_token = hparams_i.bos_token
                eos_token = hparams_i.eos_token
            bos_token = utils.default_str(bos_token, SpecialTokens.BOS)
            eos_token = utils.default_str(eos_token, SpecialTokens.EOS)

            vocab_share = hparams_i.vocab_share_with
            if vocab_share is not None:
                if vocab_share >= i:
                    MultiAlignedData._raise_sharing_error(
                        i, vocab_share, "vocab_share_with")
                if vocabs[vocab_share] is None:
                    raise ValueError(
                        f"Cannot share vocab with dataset {vocab_share} which "
                        "does not have a vocab.")
                if (bos_token == vocabs[vocab_share].bos_token and
                        eos_token == vocabs[vocab_share].eos_token):
                    vocab = vocabs[vocab_share]
                else:
                    vocab = Vocab(hparams[vocab_share].vocab_file,
                                  bos_token=bos_token, eos_token=eos_token)
            else:
                vocab = Vocab(hparams_i.vocab_file,
                              bos_token=bos_token, eos_token=eos_token)
            vocabs.append(vocab)

        return vocabs

    @staticmethod
    def make_embedding(hparams: List[HParams], vocabs: List[Optional[Vocab]]) \
            -> List[Optional[Embedding]]:
        r"""Optionally loads embeddings from files (if provided), and
        returns respective :class:`texar.torch.data.Embedding` instances.
        """
        embeddings: List[Optional[Embedding]] = []
        for i, hparams_i in enumerate(hparams):
            if not _is_text_data(hparams_i.data_type):
                embeddings.append(None)
                continue

            emb_share = hparams_i.embedding_init_share_with
            if emb_share is not None:
                if emb_share >= i:
                    MultiAlignedData._raise_sharing_error(
                        i, emb_share, "embedding_init_share_with")
                if not embeddings[emb_share]:
                    raise ValueError(
                        f"Cannot share embedding with dataset {emb_share} "
                        "which does not have an embedding.")
                if emb_share != hparams_i.vocab_share_with:
                    raise ValueError(
                        "'embedding_init_share_with' != 'vocab_share_with'."
                        "'embedding_init' can be shared only when vocab is"
                        "shared.")
                emb = embeddings[emb_share]
            else:
                emb = None
                emb_file = hparams_i.embedding_init.file
                vocab = vocabs[i]
                if emb_file and emb_file != "":
                    assert vocab is not None
                    emb = Embedding(vocab.token_to_id_map_py,
                                    hparams_i.embedding_init)
            embeddings.append(emb)

        return embeddings

    def process(self, raw_example: Tuple[Union[str, Dict[str, Any]], ...]) \
            -> Tuple[Union[List[str], int, float, Dict[str, Any]], ...]:
        processed_examples = []
        for i, raw_example_i in enumerate(raw_example):
            example_i = self._databases[i].process(raw_example_i)
            processed_examples.append(example_i)
        return tuple(processed_examples)

    def collate(self, examples) -> Batch:
        transposed_examples = map(list, zip(*examples))
        batch: Dict[str, Any] = {}
        for i, transposed_example in enumerate(transposed_examples):
            kth_batch = self._databases[i].collate(transposed_example)
            for key, name in self._names[i].items():
                batch.update({name: kth_batch[key]})
        return Batch(len(examples), batch=batch)

    def list_items(self):
        r"""Returns the list of item names that the data can produce.

        Returns:
            A list of strings.
        """
        return [value for name in self._names for value in name.values()]

    def _maybe_name_to_id(self, name_or_id):
        if is_str(name_or_id):
            if name_or_id not in self._name_to_id:
                raise ValueError("Unknown data name: {}".format(name_or_id))
            return self._name_to_id[name_or_id]
        return name_or_id

    def vocab(self, name_or_id):
        r"""Returns the :class:`~texar.torch.data.Vocab` of text dataset by its
        name or id. `None` if the dataset is not of text type.

        Args:
            name_or_id (str or int): Data name or the index of text dataset.
        """
        i = self._maybe_name_to_id(name_or_id)
        return self._vocab[i]

    def embedding_init_value(self, name_or_id):
        r"""Returns the `Tensor` of embedding initial value of the
        dataset by its name or id. `None` if the dataset is not of text type.
        """
        i = self._maybe_name_to_id(name_or_id)
        return self._embedding[i]

    def text_name(self, name_or_id):
        r"""The name of text tensor of text dataset by its name or id. If the
        dataset is not of text type, returns `None`.
        """
        i = self._maybe_name_to_id(name_or_id)
        if not _is_text_data(self._hparams.datasets[i]["data_type"]):
            return None
        return self._names[i]["text"]

    def length_name(self, name_or_id):
        r"""The name of length tensor of text dataset by its name or id. If the
        dataset is not of text type, returns `None`.
        """
        i = self._maybe_name_to_id(name_or_id)
        if not _is_text_data(self._hparams.datasets[i]["data_type"]):
            return None
        return self._names[i]["length"]

    def text_id_name(self, name_or_id):
        r"""The name of length tensor of text dataset by its name or id. If the
        dataset is not of text type, returns `None`.
        """
        i = self._maybe_name_to_id(name_or_id)
        if not _is_text_data(self._hparams.datasets[i]["data_type"]):
            return None
        return self._names[i]["text_ids"]

    def data_name(self, name_or_id):
        r"""The name of the data tensor of scalar dataset by its name or id..
        If the dataset is not a scalar data, returns `None`.
        """
        i = self._maybe_name_to_id(name_or_id)
        if not _is_scalar_data(self._hparams.datasets[i]["data_type"]):
            return None
        return self._names[i]["label"]
